"""
This module provides a plugin architecture for privacy (enryption) methods.

Each privacy plugin can be distributed as separate package by providing
modules inside the namespace-package "puresnmp_plugins.priv". Note that in
order to be a valid namespace-package, such a package *must not* have a
``__init__.py`` file!

Example folder-structure for a privacy plugin:

.. code-block:: text

    my-privacy-plugin/
     +- setup.py (or pyproject.toml)
     +- puresnmp/
         +- priv/
             +- mymodule.py
             +- myothermodule.py

Note that there is no ``__init__.py`` file!

In order for modules to be detected as plugin, they must follow the following
rules:

* Have a function ``encrypt_data`` implementing the
  :py:meth:`puresnmp_plugins.priv.TPriv.encrypt_data` protocol.
* Have a function ``decrypt_data`` implementing the
  :py:meth:`puresnmp_plugins.priv.TPriv.decrypt_data` protocol.
* Contain a string-variable ``IDENTIFIER``. This variable should be
  user-friends and is used to uniquely identify this privacy module.
"""
from types import ModuleType
from typing import NamedTuple

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


from puresnmp.exc import UnknownPrivacyModel
from puresnmp.plugins.pluginbase import Loader


class EncryptionResult(NamedTuple):
    """
    Wrapper for encrypted data.
    """

    #: The encrypted content (usually a PDU)
    encrypted_data: bytes

    #: The value that will be inserted into the "privParams" field from the
    #: SNMP spec. This should contain any "local" data that is needed to
    #: decrypt the data (like a salt for example)
    priv_params: bytes


class TPriv(Protocol):
    """
    Protocol describing the en-/de-cryption API
    """

    def encrypt_data(
        self,
        localised_key: bytes,
        engine_id: bytes,
        engine_boots: int,
        engine_time: int,
        data: bytes,
    ) -> EncryptionResult:  # pragma: no cover
        """
        Encrypts data from an SNMP PDU following the SNMPv3 spec.

        :param localised_key: The encryption key localised to the given
            engine-id
        :param engine_id: The ID of the recipient SNMP-engine
        :param engine_boots: Number of times the recipient engine has been
            restarted
        :param data: The data that needs to be encrypted
        :returns: The encrypted data together with the used salt
        """
        ...

    def decrypt_data(
        self,
        localised_key: bytes,
        engine_id: bytes,
        engine_boots: int,
        engine_time: int,
        salt: bytes,
        data: bytes,
    ) -> bytes:  # pragma: no cover
        """
        Decrypts a message

        :param localised_key: The decryption key localised to the given
            engine-id
        :param engine_id: The remote engine-id
        :param engine_boots: The number of engine-boots
        :param engine_time: Timeliness parameter for encryption
        :param salt: The salt used during encryption
        :param data: The encrypted SNMP message
        :returns: The unencrypted data.
        """
        ...


def is_valid_priv_mod(mod: ModuleType) -> bool:
    return all(
        [
            hasattr(mod, "encrypt_data"),
            hasattr(mod, "decrypt_data"),
            hasattr(mod, "IDENTIFIER"),
            hasattr(mod, "IANA_ID"),
        ]
    )


def create(identifier: str) -> TPriv:
    namespace = "puresnmp_plugins.priv"
    loader = Loader(namespace, is_valid_priv_mod)
    result = loader.create(identifier)
    if not result:
        raise UnknownPrivacyModel(
            namespace,
            identifier,
            sorted(loader.discovered_plugins.keys()),
        )
    return result  # type: ignore



"""
Example plugin for :py:mod:`puresnmp`. This does no encryption, it's only an
example.
"""
from typing import NamedTuple

IDENTIFIER = "example"
IANA_ID = -1


class EncryptionResult(NamedTuple):
    ciphertext: bytes
    salt: bytes


def encrypt_data(
    localised_key: bytes,
    engine_id: bytes,
    engine_boots: int,
    engine_time: int,
    data: bytes,
) -> EncryptionResult:
    """
    Encrypt the data
    """
    # This is where we would encrypt the data in "data"
    return EncryptionResult(data, b"")


def decrypt_data(
    localised_key: bytes,
    engine_id: bytes,
    engine_boots: int,
    engine_time: int,
    salt: bytes,
    data: bytes,
) -> bytes:
    """
    See https://tools.ietf.org/html/rfc3826#section-3.1.4
    """
    # This is where we would decrypt the data
    return data



"""
This defines a no-op security model which can be used as fallback
"""
from typing import Any, Dict

from puresnmp.plugins.security import SecurityModel

IDENTIFIER = 0


class NullSecurityModel(SecurityModel[Any, Any]):
    """
    This is a placeholder class for security models which don't implement any
    logic.

    A use-case is the "ANY" security model which is used during the discovery
    phase.
    """


def create(local_config: Dict[bytes, Dict[str, Any]]) -> NullSecurityModel:
    """
    Create a new instance of the security model
    """
    return NullSecurityModel(local_config)



"""
This file contains various values used to avoid magic numbers and strings in
the application.
"""
#: Maximum number of usable varbinds as defined in :rfc:`3416`
MAX_VARBINDS = 2147483647

#: A magic value used to detect strict error-handling
ERRORS_STRICT = "strict"

#: A magic value used to detect lenient error-handling
ERRORS_WARN = "warn"

#: TCP timeout which is used if not manually overridden
DEFAULT_TIMEOUT = 6

#: Number of retries we attempt when sending out packets before giving up
DEFAULT_RETRIES = 10

#: The default interface to use when listening to traps. This is flagged with
#: "nosec" as "secure" as this is very much intentional.
DEFAULT_LISTEN_ADDRESS = "0.0.0.0"  # nosec



"""
This module contains the high-level functions to access the library. Care is
taken to make this as pythonic as possible and hide as many of the gory
implementations as possible.
"""


from x690.types import ObjectIdentifier

# !!! DO NOT REMOVE !!! The following import triggers the processing of SNMP
# Types and thus populates the Registry. If this is not included, Non x.690
# SNMP types will not be properly detected!
import puresnmp.types
from puresnmp.api.pythonic import PyWrapper
from puresnmp.api.raw import Client
from puresnmp.credentials import V1, V2C, V3

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore


__version__ = importlib_metadata.version("puresnmp")

__all__ = [
    "Client",
    "ObjectIdentifier",
    "PyWrapper",
    "V1",
    "V2C",
    "V3",
    "__version__",
]



"""
This module provides common code for hashing based authentication.
"""
import hmac
from typing import Callable

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol


THasher = Callable[[bytes, bytes], bytes]

class TOutgoing(Protocol):
    """
    Protocol for callables that authenticate outgoing SNMP messages
    """

    # pylint: disable=too-few-public-methods

    def __call__(self, auth_key: bytes, data: bytes, engine_id: bytes) -> bytes:
        # pylint: disable=unused-argument, missing-docstring
        ...


class TIncoming(Protocol):
    """
    Protocol for callables that authenticate incoming SNMP messages
    """

    # pylint: disable=too-few-public-methods

    def __call__(
        self,
        auth_key: bytes,
        data: bytes,
        received_digest: bytes,
        engine_id: bytes,
    ) -> bool:
        # pylint: disable=unused-argument, missing-docstring
        ...


def for_outgoing(hasher: THasher, hmac_method: str) -> TOutgoing:
    """
    Create a new callable able to authenticate outgoing messages.

    :param hasher: A function used to apply a hashing algorithm.
    :param hmac_method: The specific HMAC implementation to use
    :return: A callable that can be used in a puresnmp authentication plugin
    """

    def authenticate_outgoing_message(
        auth_key: bytes, data: bytes, engine_id: bytes
    ) -> bytes:
        """
        See https://tools.ietf.org/html/rfc3414#section-1.6
        """
        digest = get_message_digest(
            hasher,
            hmac_method,
            auth_key,
            data,
            engine_id,
        )
        return digest

    return authenticate_outgoing_message


def for_incoming(hasher: THasher, hmac_method: str) -> TIncoming:
    """
    Create a new callable able to authenticate incoming messages.

    :param hasher: A function used to apply a hashing algorithm.
    :param hmac_method: The specific HMAC implementation to use
    :return: A callable that can be used in a puresnmp authentication plugin
    """

    def is_authentic(
        auth_key: bytes, data: bytes, received_digest: bytes, engine_id: bytes
    ) -> bool:
        """
        Returns whether the given data-block is authentic for a given key.

        See https://tools.ietf.org/html/rfc3414#section-1.6


        :param auth_key: The authentication key (supplied by a user)
        :param data: The data-block to verify
        :param received_digest: The "expected" digest
        :param engine_id: The SNMP engine-id for which this authentication is
            applied
        """
        expected_digest = get_message_digest(
            hasher, hmac_method, auth_key, data, engine_id
        )
        return received_digest == expected_digest

    return is_authentic


def get_message_digest(
    hasher: THasher,
    method: str,
    auth_key: bytes,
    encoded_message: bytes,
    engine_id: bytes,
) -> bytes:
    """
    Calculate the digest for a given message.

    :param hasher: A specific hash-implementation (f.ex.: ``hashlib.md5``)
    :param method: The digest-method to use
    :param auth_key: The authentication key for the user
    :param encoded_message: The SNMP message as bytes
    :param engine_id: The ID of the receiving engine
    """
    auth_key = hasher(auth_key, engine_id)
    mac = hmac.new(auth_key, encoded_message, digestmod=method)
    return mac.digest()[:12]



"""
This module contains a high-level API to SNMP functions.

The arguments and return values of these functions have types which are
internal to ``puresnmp`` (subclasses of :py:class:`x690.types.Type`).

Alternatively, there is :py:mod:`puresnmp.api.pythonic` which converts
these values into pure Python types. This makes day-to-day programming a bit
easier but loses type information which may be useful in some edge-cases.
"""

import asyncio
import logging
from asyncio import get_event_loop
from asyncio.events import AbstractEventLoop
from collections import OrderedDict
from contextlib import contextmanager
from dataclasses import dataclass, replace
from ipaddress import ip_address
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
)
from typing import Type as TType
from typing import TypeVar, cast

from x690.types import Integer, Null, ObjectIdentifier, Sequence
from x690.types import X690Type as Type

from puresnmp.plugins import mpm
from puresnmp.typevars import SocketResponse, TAnyIp

from ..const import (
    DEFAULT_LISTEN_ADDRESS,
    DEFAULT_RETRIES,
    DEFAULT_TIMEOUT,
    ERRORS_STRICT,
    ERRORS_WARN,
)
from ..credentials import V2C, Credentials
from ..exc import FaultySNMPImplementation, NoSuchOID, SnmpError
from ..pdu import (
    PDU,
    BulkGetRequest,
    EndOfMibView,
    GetNextRequest,
    GetRequest,
    NoSuchInstance,
    NoSuchObject,
    PDUContent,
    SetRequest,
    Trap,
)
from ..transport import Endpoint, TSender, listen, send_udp
from ..util import (
    BulkResult,
    TTableRow,
    get_request_id,
    get_unfinished_walk_oids,
    group_varbinds,
    tablify,
    validate_response_id,
)
from ..varbind import VarBind

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

TWalkResponse = AsyncGenerator[VarBind, None]
T = TypeVar("T", bound=TType[Any])  # pylint: disable=invalid-name

LOG = logging.getLogger(__name__)

TFetcher = Callable[[List[ObjectIdentifier]], Awaitable[List[VarBind]]]
"""
Type alias for a callable that is responsible to fetch a collection of OIDs
from the remote device
"""


def deduped_varbinds(
    requested_oids: List[ObjectIdentifier],
    grouped_oids: Dict[ObjectIdentifier, List[VarBind]],
    yielded: Set[ObjectIdentifier],
) -> Generator[VarBind, None, None]:
    """
    Generate grouped OIDs by ensuring they are contained in the original
    request and have no duplicates.

    >>> OID = ObjectIdentifier
    >>> list(deduped_varbinds(
    ...     [OID("1.2"), OID("2.3")],
    ...     {
    ...         OID("1.2"): [VarBind(OID("1.2.3.4"), 1)],
    ...         OID("1.2"): [VarBind(OID("1.2.3.4"), 1)],
    ...         OID("5.6"): [VarBind(OID("5.6.7.8"), 1)],
    ...     },
    ...     set()
    ... ))
    [VarBind(oid=ObjectIdentifier('1.2.3.4'), value=1)]

    :param requested_oids: A list of OIDs which were originally requested. If
        any value from the grouped varbinds are not children of any of these
        OIDs, the issue is logged and the value is skipped.
    :param grouped_oids: The OIDs that need to be verified & deduped.
    :param yielded: A set containing all OIDs that have already been
        generated by this function. This set will be updated by this function
        whenever a value is returned to detect duplicates.
    """
    for var in sorted(grouped_oids.values()):
        for varbind in var:
            containment = [varbind.oid in _ for _ in requested_oids]
            if not any(containment) or varbind.oid in yielded:
                LOG.debug(
                    "Unexpected device response: Returned VarBind %s "
                    "was either not contained in the requested tree or "
                    "appeared more than once. Skipping!",
                    varbind,
                )
                continue
            yielded.add(varbind.oid)
            yield varbind


@dataclass(frozen=True)
class Context:
    """
    Information about the current SNMP context
    """

    engine_id: bytes
    name: bytes


@dataclass(frozen=True)
class ClientConfig:
    """
    Overridable configuration for SNMP clients

    These settings can be overridden via :py:meth:`Client.reconfigure`
    """

    #: The credentials used to apply to SNMP requests
    credentials: Credentials
    #: The SNMPv3 Context. For SNMPv1 or SNMPv2 this value is ignored
    context: Context
    #: The SNMPv3 "local config cache". For SNMPv1 or SNMPv2 this value is
    #: ignored
    lcd: Dict[str, Any]
    #: The socket timeout for network requests. This value is passed through
    #: to the client's "sender" callable
    timeout: int = DEFAULT_TIMEOUT
    #: The number of retries we attempt when sending packets to the remote
    #: device before giving up. Note that some devices may refuse connections
    #: attempts if too many requests were made with incorrect credentials.
    #: Defaults to 10 retries
    retries: int = DEFAULT_RETRIES


class Client:
    """
    A client to execute SNMP commands on a remote device.

    To run SNMP commands on a remote device, create an instance for that
    device, and then call the instance methods.

    All functions are based on asyncio and must be used in an async context.

    Credentials need to be instances of classes taken from
    :py:mod:`puresnmp.credentials` which are used to determine the
    appropriate communication model for this client instance.

    >>> from puresnmp import Client, ObjectIdentifier, V2C
    >>> import warnings
    >>> warnings.simplefilter("ignore")
    >>> client = Client("192.0.2.1", V2C("public"))
    >>> client.get(ObjectIdentifier("1.3.6.1.2.1.1.2.0"))  # doctest: +ELLIPSIS
    <coroutine ...>

    :param ip: The IP-address of the remote SNMP device
    :param credentials: User credentials for the request. These define the
        underlying protocol in use. See :py:mod:`puresnmp.credentials` for
        possible types.
    :param port: The UDP port for the remote device
    :param sender: A callable responsible to send out data to the remote
        device. The default implementation will use UDP using the IP and port
        given in the other arguments.
    :param context_name: An optional context for SNMPv3 requests
    :param engine_id: An optional Engine ID for SNMPv3 requests. Helper
        functions are provided in :py:mod:`puresnmp.util` to generate valid IDs.
    """

    def __init__(
        self,
        ip: str,
        credentials: Credentials,
        port: int = 161,
        sender: TSender = send_udp,
        context_name: bytes = b"",
        engine_id: bytes = b"",
    ) -> None:

        lcd: Dict[str, Any] = {}
        self.config = ClientConfig(
            credentials=credentials,
            context=Context(engine_id, context_name),
            lcd=lcd,
        )

        endpoint = Endpoint(ip_address(ip), port)

        async def handler(data: bytes) -> bytes:  # pragma: no cover
            """
            A callable that is bound to a given IP/Port combination, capable of
            sending raw bytes to that endpoint. This is passed to
            message-processing models in case they need to communicate with the
            device.

            At the time of this writing, this is only required vor SNMPv3
            discovery messages.
            """
            return await sender(
                endpoint,
                data,
                timeout=self.config.timeout,
                retries=self.config.retries,
            )

        self.sender = sender
        self.transport_handler = handler
        self.endpoint = endpoint
        self.mpm = mpm.create(credentials.mpm, handler, lcd)

    @property
    def credentials(self) -> Credentials:
        """
        Accessor to the client credentials
        """
        return self.config.credentials

    @property
    def context(self) -> Context:
        """
        Accessor to the SNMPv3 context
        """
        return self.config.context

    @property
    def ip(self) -> TAnyIp:
        """
        Accessor to the endpoint IP address
        """
        return self.endpoint.ip

    @property
    def port(self) -> int:
        """
        Accessor to the endpoint port
        """
        return self.endpoint.port

    def configure(self, **kwargs: Any) -> None:
        """
        Update the configuration of the client.

        For temporary configuration changes see :py:meth:`~.reconfigure`.

        The values that can be overridden are defined in
        :py:class:`~.ClientConfig`. Any fields in that class can be overridden

        >>> client = Client("192.0.2.1", V2C("public"))
        >>> client.config.retries
        10
        >>> client.configure(retries=3)
        >>> client.config.retries
        3
        """
        new_config = replace(self.config, **kwargs)
        if "credentials" in kwargs and type(self.config.credentials) != type(
            kwargs["credentials"]
        ):
            # New credentials may switch from one SNMP version to another
            # so we need to create a new message-processing-model
            lcd: Dict[str, Any] = {}
            self.mpm = mpm.create(
                kwargs["credentials"].mpm, self.transport_handler, lcd
            )
        self.config = new_config

    @contextmanager
    def reconfigure(self, **kwargs: Any) -> Generator[None, None, None]:
        """
        Temporarily reconfigure the client.

        Some values may need to be modified during the lifetime of the client
        for some requests. A typical example would be using different
        credentials for "set" commands, or different socket timeouts for some
        targeted requests.

        This is provided via this context-manager. When the context-manager
        exits, the previous config is restored

        The values that can be overridden are defined in
        :py:class:`~.ClientConfig`. Any fields in that class can be overridden

        >>> client = Client("192.0.2.1", V2C("public"))
        >>> client.config.timeout
        6
        >>> with client.reconfigure(timeout=10):
        ...     client.config.timeout
        10
        >>> client.config.timeout
        6
        """
        old_config = self.config
        old_mpm = self.mpm
        try:
            self.configure(**kwargs)
            yield
        finally:
            self.config = old_config
            self.mpm = old_mpm

    async def _send(self, pdu: PDU, request_id: int) -> PDU:
        packet, _ = await self.mpm.encode(
            request_id,
            self.credentials,
            self.context.engine_id,
            self.context.name,
            pdu,
        )
        raw_response = await self.sender(
            self.endpoint,
            bytes(packet),
            timeout=self.config.timeout,
            retries=self.config.retries,
        )
        response = self.mpm.decode(raw_response, self.credentials)
        validate_response_id(request_id, response.value.request_id)
        return response

    async def get(self, oid: ObjectIdentifier) -> Type[Any]:
        """
        Retrieve the value of a single OID

        >>> from asyncio import run
        >>> import warnings
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> warnings.simplefilter("ignore")
        >>> client = Client("127.0.0.1", V2C("private"), port=50009)
        >>> coro = client.get(OID("1.3.6.1.2.1.1.2.0"))
        >>> run(coro)  # doctest: +SKIP
        ObjectIdentifier('1.3.6.1.4.1.8072.3.2.10')
        """
        result = await self.multiget([oid])
        if isinstance(result[0], (NoSuchObject, NoSuchInstance)):
            raise NoSuchOID(oid)
        return result[0]

    async def multiget(self, oids: List[ObjectIdentifier]) -> List[Type[Any]]:
        """
        Retrieve (scalar) values from multiple OIDs in one request.

        >>> from asyncio import run
        >>> import warnings
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> warnings.simplefilter("ignore")
        >>> client = Client("127.0.0.1", V2C("private"), port=50009)
        >>> coro = client.multiget(
        ...     [OID('1.3.6.1.2.1.1.2.0'), OID('1.3.6.1.2.1.1.1.0')]
        ... )
        >>> run(coro)  # doctest: +SKIP
        [ObjectIdentifier('1.3.6.1.4.1.8072.3.2.10'), OctetString(b'Linux c8582f39c32b 4.15.0-115-generic #116-Ubuntu SMP Wed Aug 26 14:04:49 UTC 2020 x86_64')]
        """

        parsed_oids = [VarBind(oid, Null()) for oid in oids]

        request_id = get_request_id()
        pdu = GetRequest(PDUContent(request_id, parsed_oids))
        response = await self._send(pdu, request_id)
        output = [value for _, value in response.value.varbinds]
        if len(output) != len(oids):
            raise SnmpError(
                "Unexpected response. Expected %d varbind, "
                "but got %d!" % (len(oids), len(output))
            )
        return output

    async def getnext(self, oid: ObjectIdentifier) -> VarBind:
        """
        Executes a single SNMP GETNEXT request (used inside *walk*).

        >>> from puresnmp import Client, ObjectIdentifier as OID
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> client = Client("192.0.2.1", V2C("private"))
        >>> # The line below needs to be "awaited" to get the result.
        >>> # This is not shown here to make it work with doctest
        >>> client.getnext(OID('1.2.3.4'))
        <coroutine object ...>
        """
        result = await self.multigetnext([oid])
        if isinstance(result[0], (NoSuchObject, NoSuchInstance)):
            raise NoSuchOID(oid)
        return result[0]

    async def walk(
        self,
        oid: ObjectIdentifier,
        errors: str = ERRORS_STRICT,
    ) -> TWalkResponse:
        """
        A convenience method delegating to :py:meth:`~.multiwalk` with
        exactly one OID
        """
        async for row in self.multiwalk([oid], errors=errors):
            yield row

    async def multiwalk(
        self,
        oids: List[ObjectIdentifier],
        fetcher: Optional[TFetcher] = None,
        errors: str = ERRORS_STRICT,
    ) -> TWalkResponse:
        """
        Retrieve all values "below" multiple OIDs with a single operation.

        Note: This will send out as many "GetNext" requests as needed.

        This is almost the same as :py:meth:`~.walk` except that it is
        capable of iterating over multiple OIDs at the same time.

        >>> from asyncio import run
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> async def example():
        ...     client = Client("127.0.0.1", V2C("private"), port=50009)
        ...     result = client.multiwalk(
        ...         [OID('1.3.6.1.2.1.1'), OID('1.3.6.1.4.1.1')]
        ...     )
        ...     output = []
        ...     async for row in result:
        ...         output.append(row)
        ...     return output
        >>> run(example())  # doctest: +SKIP
        [VarBind(oid=ObjectIdentifier('1.3.6.1.2.1.1.1.0'), value=Oct...]
        """
        if fetcher is None:
            fetcher = self.multigetnext

        LOG.debug("Walking on %d OIDs using %s", len(oids), fetcher.__name__)
        varbinds = await fetcher(oids)
        grouped_oids = group_varbinds(varbinds, oids)
        unfinished_oids = get_unfinished_walk_oids(grouped_oids)
        yielded: Set[ObjectIdentifier] = set()
        for varbind in deduped_varbinds(oids, grouped_oids, yielded):
            yield varbind

        # As long as we have unfinished OIDs, we need to continue the walk for
        # those.
        while unfinished_oids:
            next_fetches = [_[1].value.oid for _ in unfinished_oids]
            try:
                varbinds = await fetcher(next_fetches)
            except NoSuchOID:
                # Reached end of OID tree, finish iteration
                break
            except FaultySNMPImplementation as exc:
                if errors == ERRORS_WARN:
                    LOG.warning(
                        "SNMP walk aborted prematurely due to faulty SNMP "
                        "implementation on device %r! Upon running a "
                        "GetNext on OIDs %r it returned the following "
                        "error: %s",
                        self.endpoint,
                        next_fetches,
                        exc,
                    )
                    break
                raise
            grouped_oids = group_varbinds(
                varbinds, next_fetches, user_roots=oids
            )
            unfinished_oids = get_unfinished_walk_oids(grouped_oids)
            if LOG.isEnabledFor(logging.DEBUG) and len(oids) > 1:
                LOG.debug(
                    "%d of %d OIDs need to be continued",
                    len(unfinished_oids),
                    len(oids),
                )
            for varbind in deduped_varbinds(oids, grouped_oids, yielded):
                yield varbind

    async def multigetnext(self, oids: List[ObjectIdentifier]) -> List[VarBind]:
        """
        Executes a single multi-oid GETNEXT request.

        The request sends one packet to the remote host requesting the value
        of the OIDs following one or more given OIDs.

        >>> from asyncio import run
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> client = Client("127.0.0.1", V2C("private"), port=50009)
        >>> # The line below needs to be "awaited" to get the result.
        >>> # This is not shown here to make it work with doctest
        >>> coro = client.multigetnext(
        ...     [OID('1.3.6.1.2.1.1.2.0'), OID('1.3.6.1.2.1.1.1.0')]
        ... )
        >>> run(coro)  # doctest: +ELLIPSIS +SKIP
        [VarBind(oid=ObjectIdentifier('1.3.6.1.2.1.1.3.0'), value=TimeTicks(...)), VarBind(oid=ObjectIdentifier('1.3.6.1.2.1.1.2.0'), value=ObjectIdentifier('1.3.6.1.4.1.8072.3.2.10'))]
        """
        varbinds = [VarBind(oid, Null()) for oid in oids]
        request_id = get_request_id()
        pdu = GetNextRequest(PDUContent(request_id, varbinds))
        response_object = await self._send(pdu, request_id)
        if len(response_object.value.varbinds) != len(oids):
            raise SnmpError(
                "Invalid response! Expected exactly %d varbind, "
                "but got %d" % (len(oids), len(response_object.value.varbinds))
            )

        output = []
        for oid, value in response_object.value.varbinds:
            if isinstance(value, EndOfMibView):
                break
            output.append(VarBind(oid, value))

        # Verify that the OIDs we retrieved are successors of the requested OIDs
        for requested, retrieved in zip(oids, output):
            if not requested < retrieved.oid:
                raise FaultySNMPImplementation(
                    "The OID %s is not a successor of %s!"
                    % (retrieved.oid, requested)
                )
        return output

    async def table(
        self,
        oid: ObjectIdentifier,
        _rowtype: TType[TTableRow] = Dict[str, Any],  # type: ignore
    ) -> List[TTableRow]:
        """
        Fetch an SNMP table

        The resulting output will be a list of dictionaries where each
        dictionary corresponds to a row of the table.

        SNMP Tables are indexed as follows::

            <base-oid>.<column-id>.<row-id>

        A "row-id" can be either a single numerical value, or a partial OID.
        The row-id will be contained in key ``'0'`` of each row (as a string)
        representing that partial OID (often a suffix which can be used in
        other tables). This key ``'0'`` is automatically injected by
        ``puresnmp``. This ensures that the row-index is available even for
        tables that don't include that value themselves.

        SNMP-Tables are fetched first by column, then by row (by the nature
        of the defined MIB structure). This means that this method has to
        consume the complete table before being able to return anything.

        Example output:

        >>> from asyncio import run
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> client = Client("127.0.0.1", V2C("private"), port=50009)
        >>> coro = client.table(OID("1.3.6.1.2.1.2.2.1"))
        >>> run(coro)  # doctest: +SKIP
        [{'0': '1', '1': Integer(1), ... '22': ObjectIdentifier('0.0')}]
        """
        tmp = []
        varbinds = self.walk(oid)
        async for varbind in varbinds:
            tmp.append(varbind)
        as_table: List[TTableRow] = tablify(
            tmp, num_base_nodes=len(oid), _rowtype=_rowtype
        )
        return as_table

    async def set(
        self,
        oid: ObjectIdentifier,
        value: T,
    ) -> T:
        """
        Update a value on the remote host

        Values must be a subclass of :py:class:`x690.types.Type`. See
        :py:mod:`x690.types` for a predefined collection of types.

        >>> from asyncio import run
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> from x690.types import OctetString
        >>> client = Client("127.0.0.1", V2C("private"), port=50009)
        >>> coro = client.set(
        ...     OID("1.3.6.1.2.1.1.4.0"), OctetString(b'new contact value')
        ... )
        >>> run(coro)  # doctest: +SKIP
        OctetString(b'new contact value')
        """
        value_internal = cast(Type[Any], value)
        result = await self.multiset({oid: value_internal})
        return result[oid]  # type: ignore

    async def multiset(
        self, mappings: Dict[ObjectIdentifier, Type[Any]]
    ) -> Dict[ObjectIdentifier, Type[Any]]:
        """
        Executes an SNMP SET request on multiple OIDs. The result is returned as
        pure Python data structure.

        >>> from asyncio import run
        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> from x690.types import OctetString
        >>> client = Client("127.0.0.1", V2C("private"), port=50009)
        >>> coro = client.multiset({
        ...     OID('1.3.6.1.2.1.1.4.0'): OctetString(b'new-contact'),
        ...     OID('1.3.6.1.2.1.1.6.0'): OctetString(b'new-location')
        ... })
        >>> run(coro)  # doctest: +ELLIPSIS +SKIP
        {ObjectIdentifier('1.3.6.1.2.1.1.4.0'): OctetString(b'new-c...cation')}
        """

        if any(not isinstance(v, Type) for v in mappings.values()):
            raise TypeError(
                "SNMP requires typing information. The value for a "
                '"set" request must be an instance of "Type"!'
            )

        binds = [VarBind(oid, value) for oid, value in mappings.items()]

        pdu = SetRequest(PDUContent(get_request_id(), binds))
        response = await self._send(pdu, get_request_id())

        output = dict(response.value.varbinds)
        if len(output) != len(mappings):
            raise SnmpError(
                "Unexpected response. Expected %d varbinds, "
                "but got %d!" % (len(mappings), len(output))
            )
        return output

    async def bulkget(
        self,
        scalar_oids: List[ObjectIdentifier],
        repeating_oids: List[ObjectIdentifier],
        max_list_size: int = 1,
    ) -> BulkResult:
        # pylint: disable=unused-argument, too-many-locals
        """
        Runs a "bulk" get operation and returns a :py:class:`~.BulkResult`
        instance. This contains both a mapping for the scalar variables (the
        "non-repeaters") and an OrderedDict instance containing the remaining
        list (the "repeaters").

        The OrderedDict is ordered the same way as the SNMP response
        (whatever the remote device returns).

        This operation can retrieve both single/scalar values *and* lists of
        values ("repeating values") in one single request. You can for
        example retrieve the hostname (a scalar value), the list of
        interfaces (a repeating value) and the list of physical entities
        (another repeating value) in one single request.

        Note that this behaves like a **getnext** request for scalar values!
        So you will receive the value of the OID which is *immediately
        following* the OID you specified for both scalar and repeating
        values!

        :param scalar_oids: contains the OIDs that should be fetched as single
            value.
        :param repeating_oids: contains the OIDs that should be fetched as list.
        :param max_list_size: defines the max length of each list.

        >>> from puresnmp import Client, ObjectIdentifier as OID, V2C
        >>> import warnings
        >>> warnings.simplefilter("ignore")
        >>> client = Client("192.0.2.1", V2C("private"), port=50009)
        >>> result = client.bulkget(  # doctest: +SKIP
        ...     scalar_oids=[
        ...         OID('1.3.6.1.2.1.1.1'),
        ...         OID('1.3.6.1.2.1.1.2'),
        ...     ],
        ...     repeating_oids=[
        ...         OID('1.3.6.1.2.1.3.1'),
        ...         OID('1.3.6.1.2.1.5.1'),
        ...     ],
        ...     max_list_size=10
        ... )
        BulkResult(
            scalars={
                ObjectIdentifier('1.3.6.1.2.1.1.1.0'): OctetString(
                    b'Linux c8582f39c32b 4.15.0-115-generic #116-Ubuntu SMP '
                    b'Wed Aug 26 14:04:49 UTC 2020 x86_64'
                ),
                ObjectIdentifier('1.3.6.1.2.1.1.2.0'): ObjectIdentifier(
                    '1.3.6.1.4.1.8072.3.2.10'
                )
            },
            listing=OrderedDict([
                (
                    ObjectIdentifier('1.3.6.1.2.1.3.1.1.1.8769.1.10.100.0.1'),
                    Integer(8769),
                ),
                (ObjectIdentifier('1.3.6.1.2.1.5.1.0'), Counter(1)),
                (
                    ObjectIdentifier('1.3.6.1.2.1.3.1.1.2.8769.1.10.100.0.1'),
                    OctetString(b'\x02B\x03\x96#>'),
                ),
                (ObjectIdentifier('1.3.6.1.2.1.5.2.0'), Counter(0)),
                (
                    ObjectIdentifier('1.3.6.1.2.1.3.1.1.3.8769.1.10.100.0.1'),
                    IpAddress(IPv4Address('10.100.0.1')),
                ),
                (ObjectIdentifier('1.3.6.1.2.1.5.3.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.4.1.0'), Integer(1)),
                (ObjectIdentifier('1.3.6.1.2.1.5.4.0'), Counter(1)),
                (ObjectIdentifier('1.3.6.1.2.1.4.2.0'), Integer(64)),
                (ObjectIdentifier('1.3.6.1.2.1.5.5.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.4.3.0'), Counter(4)),
                (ObjectIdentifier('1.3.6.1.2.1.5.6.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.4.4.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.5.7.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.4.5.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.5.8.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.4.6.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.5.9.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.4.7.0'), Counter(0)),
                (ObjectIdentifier('1.3.6.1.2.1.5.10.0'), Counter(0))
            ])
        )
        """

        scalar_oids = scalar_oids or []  # protect against empty values
        repeating_oids = repeating_oids or []  # protect against empty values

        oids = list(scalar_oids) + list(repeating_oids)

        non_repeaters = len(scalar_oids)

        request_id = get_request_id()
        pdu = BulkGetRequest(request_id, non_repeaters, max_list_size, *oids)
        get_response = await self._send(pdu, request_id)

        # See RFC=3416 for details of the following calculation
        n = min(non_repeaters, len(oids))
        m = max_list_size
        r = max(len(oids) - n, 0)  # pylint: disable=invalid-name
        expected_max_varbinds = n + (m * r)

        n_retrieved_varbinds = len(get_response.value.varbinds)
        if n_retrieved_varbinds > expected_max_varbinds:
            raise SnmpError(
                "Unexpected response. Expected no more than %d "
                "varbinds, but got %d!"
                % (expected_max_varbinds, n_retrieved_varbinds)
            )

        # cut off the scalar OIDs from the listing(s)
        scalar_tmp = get_response.value.varbinds[0 : len(scalar_oids)]
        repeating_tmp = get_response.value.varbinds[len(scalar_oids) :]

        # prepare output for scalar OIDs
        scalar_out = dict(scalar_tmp)

        # prepare output for listing
        repeating_out = OrderedDict()  # type: Dict[ObjectIdentifier, Type[Any]]
        for oid, value in repeating_tmp:
            if isinstance(value, EndOfMibView):
                break
            repeating_out[oid] = value

        return BulkResult(scalar_out, repeating_out)

    def _bulkwalk_fetcher(self, bulk_size: int = 10) -> TFetcher:
        """
        Create a bulk fetcher with a fixed limit on "repeatable" OIDs.
        """

        async def fetcher(oids: List[ObjectIdentifier]) -> List[VarBind]:
            """
            Executes a SNMP BulkGet request.
            """
            result = await self.bulkget([], oids, max_list_size=bulk_size)
            return [VarBind((k), v) for k, v in result.listing.items()]

        fetcher.__name__ = "_bulkwalk_fetcher(%d)" % bulk_size
        return fetcher

    async def bulkwalk(
        self,
        oids: List[ObjectIdentifier],
        bulk_size: int = 10,
    ) -> TWalkResponse:
        """
        Identical to :py:meth:`~.walk` but uses "bulk" requests instead.

        "Bulk" requests fetch more than one OID in one request, so they are
        more efficient, but large return-values may overflow the transport
        buffer.

        :param oids: Delegated to :py:meth:`~.walk`
        :param bulk_size: Number of values to fetch per request.
        """

        if not isinstance(oids, list):
            raise TypeError("OIDS need to be passed as list!")

        result = self.multiwalk(
            oids,
            fetcher=self._bulkwalk_fetcher(bulk_size),
        )
        async for oid, value in result:
            yield VarBind(oid, value)

    async def bulktable(
        self,
        oid: ObjectIdentifier,
        bulk_size: int = 10,
        _rowtype: TType[TTableRow] = Dict[str, Any],  # type: ignore
    ) -> List[TTableRow]:
        """
        Identical to :py:meth:`~.table` but uses "bulk" requests.

        "Bulk" requests fetch more than one OID in one request, so they are
        more efficient, but large return-values may overflow the transport
        buffer.

        :param oid: Delegated to :py:meth:`~.table`
        :param bulk_size: Number of values to fetch per request.
        """
        tmp = []
        varbinds = self.bulkwalk([oid], bulk_size=bulk_size)
        async for varbind in varbinds:
            tmp.append(varbind)
        as_table = tablify(tmp, num_base_nodes=len(oid) + 1, _rowtype=_rowtype)
        return as_table


def register_trap_callback(
    callback: Callable[[PDU], Any],
    listen_address: str = DEFAULT_LISTEN_ADDRESS,
    port: int = 162,
    credentials: Credentials = V2C("public"),
    loop: Optional[AbstractEventLoop] = None,
) -> AbstractEventLoop:
    """
    Registers a callback function for for SNMP traps.

    Every time a trap is received, the callback is called with the PDU
    contained in that trap.

    As per :rfc:`3416#section-4.2.6`, the first two varbinds are the system
    uptime and the trap OID. The following varbinds are the body of the trap

    The callback will be called on the current asyncio loop. Alternatively, a
    loop can be passed into this function in which case, the traps will be
    handler on that loop instead.
    """
    if loop is None:
        loop = get_event_loop()

    def decode(packet: SocketResponse) -> None:
        async def handler(data: bytes) -> bytes:
            return await send_udp(
                Endpoint(ip_address(packet.info.address), packet.info.port),
                data,
            )

        lcd: Dict[str, Any] = {}

        as_sequence = Sequence.decode(packet.data)

        obj = cast(Tuple[Integer, Integer, Trap], as_sequence[0])

        mproc = mpm.create(obj[0].value, handler, lcd)
        trap = mproc.decode(packet.data, credentials)
        asyncio.ensure_future(callback(trap))

    handler = listen(listen_address, port, decode, loop)
    loop.run_until_complete(handler)
    return loop



"""
Models for SNMP PDUs (Request/Response messages).

PDUs all have a common structure, which is handled in the
:py:class:`~.PDU` class. The different (basic) PDU types only differ in
their type identifier header (f.ex. ``b'\\xa0'`` for a
:py:class:`~.GetRequest`).

A SNMP PDU contains the following elements:

request-id
    A unique ID used to match requests with responses. Each reques/response
    pair share this value.

error-status
    An integer defining an error-state. In :py:mod:`puresnmp` they are mapped
    to exceptions of the class :py:exc:`puresnmp.exc.ErrorResponse`.

error-index
    If applicable, this identifies the Request OID that caused the error
    (1-indexed.)

varbinds
    A key/value pair representing the payload of the PDU. For requests, the
    value should be :py:class:`x690.types.Null`
"""

import logging
from dataclasses import dataclass
from typing import Any, List, Optional, Union, cast

from x690 import decode
from x690.types import (
    _SENTINEL_UNINITIALISED,
    UNINITIALISED,
    Integer,
    Null,
    ObjectIdentifier,
    Sequence,
    TWrappedPyType,
)
from x690.types import X690Type as Type
from x690.util import TypeClass, TypeInfo, TypeNature, encode_length

from .const import MAX_VARBINDS
from .exc import EmptyMessage, ErrorResponse, TooManyVarbinds
from .typevars import SocketInfo
from .varbind import VarBind

LOG = logging.getLogger(__name__)


@dataclass(frozen=True)
class PDUContent:
    """
    A helper class to wrap PDU data into a single "value" variable for x.690
    types.
    """

    request_id: int
    varbinds: List[VarBind]
    error_status: int = 0
    error_index: int = 0


class PDU(Type[PDUContent]):
    """
    The superclass for SNMP Messages (GET, SET, GETNEXT, ...)
    """

    #: The typeclass identifire for type-detection in :py:mod:`x690`
    TYPECLASS = TypeClass.CONTEXT

    #: The tag for type-detection in :py:mod:`x690`
    TAG = 0

    @classmethod
    def decode_raw(cls, data: bytes, slc: slice = slice(None)) -> PDUContent:
        """
        This method takes a :py:class:`bytes` object and converts it to
        an application object. This is callable from each subclass of
        :py:class:`~.PDU`.
        """
        if not data:
            raise EmptyMessage("No data to decode!")
        request_id, nxt = decode(data, slc.start or 0, enforce_type=Integer)
        error_status, nxt = decode(data, nxt, enforce_type=Integer)
        error_index, nxt = decode(data, nxt, enforce_type=Integer)

        if error_status.value:
            error_detail, nxt = decode(data, nxt, enforce_type=Sequence)
            varbinds = [VarBind(oid, value) for oid, value in error_detail]  # type: ignore
            offending_oid = None
            if error_index.value != 0:
                offending_oid = varbinds[error_index.value - 1].oid
            exception = ErrorResponse.construct(
                error_status.value, offending_oid or ObjectIdentifier()
            )
            raise exception

        values, nxt = decode(data, nxt, enforce_type=Sequence)

        if not isinstance(values, Sequence):
            raise TypeError(
                "PDUs can only be decoded from sequences but got "
                "%r instead" % type(values)
            )

        varbinds = []
        for oid, value in values:  # type: ignore
            oid = cast(ObjectIdentifier, oid)  # type: ignore
            value = cast(Type[Any], value)  # type: ignore
            varbinds.append(VarBind(oid, value))

        return PDUContent(
            request_id.value, varbinds, error_status.value, error_index.value
        )

    def encode_raw(self) -> bytes:
        """
        Encodes this instance into raw x.690 bytes (excluding type & lenght)
        """

        wrapped_varbinds = [
            Sequence([vb.oid, vb.value]) for vb in self.value.varbinds
        ]
        data: List[Type[Any]] = [
            Integer(self.value.request_id),
            Integer(self.value.error_status),
            Integer(self.value.error_index),
            Sequence(wrapped_varbinds),  # type: ignore
        ]
        payload = b"".join([bytes(chunk) for chunk in data])
        return payload

    def __repr__(self) -> str:
        try:
            return "%s(%r, %r)" % (
                self.__class__.__name__,
                self.value.request_id,
                self.value.varbinds,
            )
        except:  # pylint: disable=bare-except
            LOG.exception(
                "Exception caught in __repr__ of %s", self.__class__.__name__
            )
            return f"<{self.__class__.__name__} (error-in repr)>"

    def __eq__(self, other: Any) -> bool:
        # pylint: disable=unidiomatic-typecheck
        return type(other) == type(self) and self.value == other.value

    def pretty(self, depth: int = 0) -> str:  # pragma: no cover
        """
        Returns a "prettified" string representing the SNMP message.
        """
        prefix = "  " * depth
        lines = [
            f"{prefix}{self.__class__.__name__} (tag: {self.TAG})",
            f"{prefix}  Request ID: {self.value.request_id}",
            f"{prefix}  Error Status: {self.value.error_status}",
            f"{prefix}  Error Index: {self.value.error_index}",
        ]
        if self.value.varbinds:
            lines.append(f"{prefix}  Varbinds: ")
            for bind in self.value.varbinds:
                lines.append(f"{prefix}    {bind.oid}: {bind.value}")
        else:
            lines.append(f"{prefix}  Varbinds: <none>")

        return "\n".join(lines)


class NoSuchObject(Type[None]):
    """
    Sentinel value to detect noSuchObject
    """

    # pylint: disable=too-few-public-methods
    # |
    # | This class make exclusive use of the parent-implementation, only
    # | modifying class-level "type-detection" variables

    # This subclassesPDU for type-consistency
    TYPECLASS = TypeClass.CONTEXT
    NATURE = [TypeNature.PRIMITIVE]
    TAG = 0

    def __init__(
        self,
        value: Union[TWrappedPyType, _SENTINEL_UNINITIALISED] = UNINITIALISED,
    ) -> None:
        if value is UNINITIALISED:
            super().__init__(value=None)
        else:
            super().__init__(value=value)


class NoSuchInstance(Type[None]):
    """
    Sentinel value to detect noSuchInstance
    """

    # pylint: disable=too-few-public-methods
    # |
    # | This class make exclusive use of the parent-implementation, only
    # | modifying class-level "type-detection" variables

    # This subclassesPDU for type-consistency
    TYPECLASS = TypeClass.CONTEXT
    NATURE = [TypeNature.PRIMITIVE]
    TAG = 1

    def __init__(
        self,
        value: Union[TWrappedPyType, _SENTINEL_UNINITIALISED] = UNINITIALISED,
    ) -> None:
        if value is UNINITIALISED:
            super().__init__(value=None)
        else:
            super().__init__(value=value)


class EndOfMibView(Type[None]):
    """
    Sentinel value to detect the SNMP ``endOfMibView`` marker
    """

    # pylint: disable=too-few-public-methods
    # |
    # | This class make exclusive use of the parent-implementation, only
    # | modifying class-level "type-detection" variables

    # This subclassesPDU for type-consistency
    TYPECLASS = TypeClass.CONTEXT
    NATURE = [TypeNature.PRIMITIVE]
    TAG = 2

    def __init__(
        self,
        value: Union[TWrappedPyType, _SENTINEL_UNINITIALISED] = UNINITIALISED,
    ) -> None:
        if value is UNINITIALISED:
            super().__init__(value=None)
        else:
            super().__init__(value=value)


class GetRequest(PDU):
    """
    Represents an SNMP Get Request.
    """

    TAG = 0


class GetResponse(PDU):
    """
    Represents an SNMP basic response (this may be returned for other requests
    than GET as well).
    """

    TAG = 2


class GetNextRequest(GetRequest):
    """
    Represents an SNMP GetNext Request.
    """

    TAG = 1


class SetRequest(PDU):
    """
    Represents an SNMP SET Request.
    """

    TAG = 3


class BulkGetRequest(PDU):
    """
    Represents a SNMP GetBulk request
    """

    # pylint: disable=abstract-method

    TYPECLASS = TypeClass.CONTEXT
    TAG = 5

    def __init__(self, request_id, non_repeaters, max_repeaters, *oids):
        # type: (int, int, int, ObjectIdentifier) -> None
        if len(oids) > MAX_VARBINDS:
            raise TooManyVarbinds(len(oids))
        self.request_id = request_id
        self.non_repeaters = non_repeaters
        self.max_repeaters = max_repeaters
        self.varbinds = []  # type: List[VarBind]
        for oid in oids:
            self.varbinds.append(VarBind(oid, Null()))

    def __bytes__(self) -> bytes:
        wrapped_varbinds = [
            Sequence([vb.oid, vb.value]) for vb in self.varbinds
        ]
        data: List[Type[Any]] = [
            Integer(self.request_id),
            Integer(self.non_repeaters),
            Integer(self.max_repeaters),
            Sequence(wrapped_varbinds),  # type: ignore
        ]
        payload = b"".join([bytes(chunk) for chunk in data])

        tinfo = TypeInfo(TypeClass.CONTEXT, TypeNature.CONSTRUCTED, self.TAG)
        length = encode_length(len(payload))
        return bytes(tinfo) + length + payload

    def __repr__(self) -> str:
        try:
            oids = [repr(oid) for oid, _ in self.varbinds]
            return "%s(%r, %r, %r, %s)" % (
                self.__class__.__name__,
                self.request_id,
                self.non_repeaters,
                self.max_repeaters,
                ", ".join(oids),
            )
        except:  # pylint: disable=bare-except
            LOG.exception(
                "Exception caught in __repr__ of %s", self.__class__.__name__
            )
            return f"<{self.__class__.__name__} (error-in repr)>"

    def __eq__(self, other):
        # type: (Any) -> bool
        # pylint: disable=unidiomatic-typecheck
        return (
            type(other) == type(self)
            and self.request_id == other.request_id
            and self.non_repeaters == other.non_repeaters
            and self.max_repeaters == other.max_repeaters
            and self.varbinds == other.varbinds
        )

    def pretty(self, depth: int = 0) -> str:  # pragma: no cover
        """
        Returns a "prettified" string representing the SNMP message.
        """
        prefix = "  " * depth
        lines = [
            f"{prefix}{self.__class__.__name__} (tag: {self.TAG})",
            f"{prefix}  Request ID: {self.request_id}",
            f"{prefix}  Non Repeaters: {self.non_repeaters}",
            f"{prefix}  Max Repeaters: {self.max_repeaters}",
            f"{prefix}  Varbinds: ",
        ]
        if self.varbinds:
            lines.append(f"{prefix}  Varbinds: ")
            for bind in self.varbinds:
                lines.append(f"{prefix}    {bind.oid}: {bind.value}")
        else:
            lines.append(f"{prefix}  Varbinds: <none>")

        return "\n".join(lines)


class InformRequest(PDU):
    """
    Represents an SNMP Inform request
    """

    TAG = 6


class Trap(PDU):
    """
    Represents an SNMP Trap
    """

    TAG = 7

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super().__init__(*args, **kwargs)
        self.source = None  # type: Optional[SocketInfo]


class Report(PDU):
    """
    Represents an SNMP report
    """

    TAG = 8



"""
This module provides the plugin for MD5-based authentication
"""
import hashlib

import puresnmp_plugins.auth.hashbase as hashbase
from puresnmp.util import password_to_key

IDENTIFIER = "md5"
IANA_ID = 2

hasher = password_to_key(hashlib.md5, 16)

#: Compare incoming message digest with expected value. Return True if the
#: digest matches the expected value.
authenticate_incoming_message = hashbase.for_incoming(hasher, "md5")

#: Calculate the message digest for a SNMPv3 message.
authenticate_outgoing_message = hashbase.for_outgoing(hasher, "md5")



"""
SMI Types / Structure types which are not defined in the x690 protocol (see
also :py:mod:`x690`).

See `RFC 1155 section 3.2.3`_ for a description of the types and `RFC 3416`_
for the definition of the new types.

.. note::
    The IPv6 type is not defined in the default RFCs and needs to be
    post-processed.

.. _RFC 1155 section 3.2.3: https://tools.ietf.org/html/rfc1155#section-3.2.3
.. _RFC 3416: https://tools.ietf.org/html/rfc3416
"""
# TODO: Implement IPv6 via https://tools.ietf.org/html/rfc2465
# pylint: disable=too-few-public-methods

from datetime import timedelta
from ipaddress import IPv4Address, ip_address
from typing import Optional, Union

from x690.types import (
    _SENTINEL_UNINITIALISED,
    UNINITIALISED,
    Integer,
    OctetString,
)
from x690.types import X690Type as Type
from x690.util import TypeClass


class IpAddress(Type[IPv4Address]):
    """
    SNMP Type for IPv4 Addresses
    """

    NATURE = OctetString.NATURE
    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x00

    def encode_raw(self) -> bytes:
        """
        Converts ip-address instance into raw bytes

        >>> IpAddress(ip_address('192.0.2.1')).encode_raw()
        b'\\xc0\\x00\\x02\\x01'
        """
        numeric = int(self.value)
        return numeric.to_bytes(4, "big")

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> IPv4Address:
        """
        Converts raw-bytes to an ip-address instance

        >>> IpAddress.decode_raw(b"\\xc0\\x00\\x02\\x01")
        IPv4Address('192.0.2.1')
        """
        value = ip_address(int.from_bytes(data[slc], "big"))
        return value  # type: ignore

    def __eq__(self, other: object) -> bool:
        # TODO: no longer necessary in x690 > 0.5.0a4
        return isinstance(other, IpAddress) and self.value == other.value


class Counter(Integer):
    """
    SNMP type for counters.
    """

    SIGNED = False
    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x01

    def __init__(
        self, value: Union[int, _SENTINEL_UNINITIALISED] = UNINITIALISED
    ) -> None:
        if not isinstance(value, _SENTINEL_UNINITIALISED):
            value &= 0xFFFFFFFF if value >= 2**32 else value
            if value <= 0:
                value = 0
        super().__init__(value)


class Gauge(Integer):
    """
    SNMP type for gauges.
    """

    SIGNED = False
    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x02


class TimeTicks(Integer):
    """
    SNMP type for time ticks exposed as Python :py:class:`datetime.timedelta`
    """

    SIGNED = False
    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x03

    def __init__(
        self,
        value: Union[timedelta, int, _SENTINEL_UNINITIALISED] = UNINITIALISED,
    ) -> None:
        if isinstance(value, timedelta):
            value = int(value.total_seconds() * 100)
        super().__init__(value)

    def pythonize(self) -> Optional[timedelta]:  # type: ignore
        """
        Convert to Python type
        """
        if self.value is None:
            return None
        seconds = self.value / 100.0  # see rfc2578#section-7.1.8
        return timedelta(seconds=seconds)


class Opaque(OctetString):
    """
    The Opaque type is to be considered to carry "any" binary data.

    It is up to the application to know how to interpret this data and is
    passed through transparently by the SNMP protocol.
    """

    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x04


class NsapAddress(Integer):
    """
    Wrapped type for an NSAP Address
    """

    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x05


class Counter64(Integer):
    """
    As defined in RFC 2578
    """

    SIGNED = False
    TYPECLASS = TypeClass.APPLICATION
    TAG = 0x06

    def __init__(
        self, value: Union[int, _SENTINEL_UNINITIALISED] = UNINITIALISED
    ) -> None:
        if not isinstance(value, _SENTINEL_UNINITIALISED):
            value &= 0xFFFFFFFFFFFFFFFF if value >= 2**64 else value
            if value <= 0:
                value = 0
        super().__init__(value)


def _walk_subclasses(cls, indent=0):  # pragma: no cover
    # type: (type, int) -> None
    """
    Recursively walk over the :py:class:`Type` hierarchy and print out ReST
    formatted text on stdout.
    """
    if cls.__module__ == "__main__":
        modname = "puresnmp.types"
    else:
        modname = cls.__module__

    cname = ".".join([modname, cls.__qualname__])
    ref = ":py:class:`%s`" % cname

    print("\n", "   " * indent, "* ", ref)
    for subclass in sorted(
        cls.__subclasses__(), key=lambda x: x.__module__ + x.__name__
    ):
        _walk_subclasses(subclass, indent + 1)


def main():  # pragma: no cover
    # type: () -> int
    """
    Entrypoint for::

        python -m puresnmp.types

    This will output a RST formatted document containing the available types.
    This function was written to generate a documentation page with the
    available types.
    """

    print(".. _type_tree:\n")
    print("Type Tree")
    print("=========\n")
    _walk_subclasses(Type)
    return 0


if __name__ == "__main__":  # pragma: no cover
    import sys  # pylint: disable=ungrouped-imports

    sys.exit(main())



"""
This module provides the plugin for SNMPv1 message processing
"""
from typing import Any, Awaitable, Callable, Dict, Optional
from warnings import warn

from x690 import decode
from x690.types import Sequence

from puresnmp.credentials import V1, Credentials
from puresnmp.pdu import PDU
from puresnmp.plugins.mpm import AbstractEncodingResult, MessageProcessingModel
from puresnmp.plugins.security import SecurityModel
from puresnmp.plugins.security import create as create_sm

IDENTIFIER = 0


TV1SecModel = SecurityModel[PDU, Sequence]


class V1EncodingResult(AbstractEncodingResult):
    data: bytes
    security_model: Optional[TV1SecModel] = None


class V1MPM(MessageProcessingModel[V1EncodingResult, TV1SecModel]):
    """
    Message Processing Model for SNMPv1
    """

    async def encode(
        self,
        request_id: int,
        credentials: Credentials,
        engine_id: bytes,
        context_name: bytes,
        pdu: PDU,
    ) -> V1EncodingResult:
        if not isinstance(credentials, V1):
            raise TypeError("SNMPv1 MPM should be used with V1 credentials!")

        # The SNMPv1 support is blatantly ripping off the SNMPv2c
        # implementation which is *mostly* equivalent. Once the distinctions
        # have been ironed out, this warning can be removed.
        warn("Experimental SNMPv1 support", UserWarning)

        security_model_id = 1
        if self.security_model is None:
            self.security_model = create_sm(security_model_id)

        packet = self.security_model.generate_request_message(
            pdu, b"", credentials
        )

        return V1EncodingResult(bytes(packet))

    def decode(
        self,
        whole_msg: bytes,
        credentials: Credentials,
    ) -> PDU:
        """
        Convert an SNMPv2c message into a PDU instance
        """

        # The SNMPv1 support is blatantly ripping off the SNMPv2c
        # implementation which is *mostly* equivalent. Once the distinctions
        # have been ironed out, this warning can be removed.
        warn("Experimental SNMPv1 support", UserWarning)

        security_model_id = 1
        if self.security_model is None:
            self.security_model = create_sm(security_model_id)

        decoded, _ = decode(whole_msg, enforce_type=Sequence)
        _, _, pdu = decoded

        # Because PDUs are lazy, we need to trigger the readout of the PDU
        # value. Otherwise, any error-response is hidden, causing cryptic
        # errors.
        pdu.value

        msg = self.security_model.process_incoming_message(decoded, credentials)
        return msg


def create(
    transport_handler: Callable[[bytes], Awaitable[bytes]],
    lcd: Dict[str, Any],
) -> MessageProcessingModel[V1EncodingResult, TV1SecModel]:
    """
    Creates a new instance for V1 message processing
    """
    return V1MPM(transport_handler, lcd)



"""
This package provides two sub-packages:

* :py:mod:`puresnmp.api.pythonic`
  For most applications this is a convenient entry-point.

  A thin wrapper around the "raw" API. This wrapper decouples the code using
  this library from internal data-types. This makes the code more robust
  against internal API changes at the cost of some flexibility and
  performance (more conversions take place)

* :py:mod:`puresnmp.api.raw`

  The underlying lower-level implementation. This layer works primarily with
  internal data-types. It is still high-level enough for most applications.
  But the usage of internal data-types make this less robust against future
  updates.
"""



"""
Exceptions for the puresnmp package.

Most exceptions in this module are based on :rfc:`3416`
"""

from typing import Any, List

from x690.types import ObjectIdentifier

from puresnmp.const import MAX_VARBINDS


class SnmpError(Exception):
    """
    Generic exception originating from the puresnmp package. Every SNMP related
    error inherits from this class.
    """

    # pylint: disable=too-few-public-methods


class ErrorResponse(SnmpError):
    """
    A superclass used when the SNMP agent responded with additional error
    information.
    """

    #: Default message to report for this error (if not overridden)
    DEFAULT_MESSAGE: str = "unknown error"

    #: The "error-status" value
    IDENTIFIER: int = 0

    #: the raw (int) value of the error-status as returned by the SNMP agent.
    error_status: int

    #: the OID identified in the error message which caused the error.
    offending_oid: ObjectIdentifier

    @staticmethod
    def construct(
        error_status: int, offending_oid: ObjectIdentifier, message: str = ""
    ) -> "ErrorResponse":
        """
        Creates a new instance of an ErrorResponse class, using the proper
        subclass for the given *error_status* value. The message is optional,
        and if not specified, will use the default message for the given class.
        """
        classes = {
            cls.IDENTIFIER: cls for cls in ErrorResponse.__subclasses__()
        }
        if error_status in classes:
            cls = classes[error_status]
            return cls(offending_oid, message)
        return ErrorResponse(offending_oid, message, error_status=error_status)

    def __init__(
        self,
        offending_oid: ObjectIdentifier,
        message: str = "",
        error_status: int = 0,
    ) -> None:
        error_status = error_status or self.IDENTIFIER
        super().__init__(
            "%s (status-code: %r) on OID %s"
            % (
                message or self.DEFAULT_MESSAGE,
                error_status,
                offending_oid or "unknown",
            )
        )
        self.error_status = error_status
        self.offending_oid = offending_oid


class TooBig(ErrorResponse):
    """
    This error is returned whenever the size of the generatred response exceeds
    a size-limit defined by the queried device.
    """

    DEFAULT_MESSAGE = "SNMP response was too big!"
    IDENTIFIER = 1


class NoSuchOID(ErrorResponse):
    """
    This error is returned in any of the following cases:

    * the targeted OID does not support "snmp-set" operations
    * the targeted OID does not exist.
    * the targeted OID is an SMI aggregate type.
    * the targeted OID does not precede a known name in the MIB view.
    """

    DEFAULT_MESSAGE = "No such name/oid"
    IDENTIFIER = 2


class BadValue(ErrorResponse):
    """
    This error is returned whenever a variable is set using an incompatible
    type.
    """

    DEFAULT_MESSAGE = "Bad value"
    IDENTIFIER = 3


class ReadOnly(ErrorResponse):
    """
    This error is returned whenever a variable is set which is not writable.
    """

    DEFAULT_MESSAGE = "Value is read-only!"
    IDENTIFIER = 4


class GenErr(ErrorResponse):
    """
    This error is returned for any error which is not covered in the previous
    error classes.
    """

    DEFAULT_MESSAGE = "General Error (genErr)"
    IDENTIFIER = 5


class NoAccess(ErrorResponse):
    """
    This error is returned whenever .
    """

    DEFAULT_MESSAGE = "No Access!"
    IDENTIFIER = 6


class WrongType(ErrorResponse):
    IDENTIFIER = 7


class WrongLength(ErrorResponse):
    IDENTIFIER = 8


class WrongEncoding(ErrorResponse):
    IDENTIFIER = 9


class WrongValue(ErrorResponse):
    IDENTIFIER = 10


class NoCreation(ErrorResponse):
    IDENTIFIER = 11


class InconsistentValue(ErrorResponse):
    IDENTIFIER = 12


class ResourceUnavailable(ErrorResponse):
    IDENTIFIER = 13


class CommitFailed(ErrorResponse):
    IDENTIFIER = 14


class UndoFailed(ErrorResponse):
    IDENTIFIER = 15


class AuthorizationError(ErrorResponse):
    IDENTIFIER = 16


class NotWritable(ErrorResponse):
    IDENTIFIER = 17


class InconsistentName(ErrorResponse):
    IDENTIFIER = 18


class EmptyMessage(SnmpError):
    """
    Raised when trying to decode an SNMP-Message with no content.
    """


class TooManyVarbinds(SnmpError):
    """
    Exception which is raised when the number of VarBinds exceeds the limit
    defined in :rfc:`3416`.
    """

    def __init__(self, num_oids):
        # type: (int) -> None
        super().__init__(
            "Too many VarBinds (%d) in one request!"
            " RFC3416 limits requests to %d!" % (num_oids, MAX_VARBINDS)
        )
        self.num_oids = num_oids


class Timeout(SnmpError):
    """
    Wrapper for network timeouts.

    This wraps both "socket.timeout" and "asyncio.TimeoutError"
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class FaultySNMPImplementation(SnmpError):
    """
    Exception which indicates an unexpected response from an SNMP agent.
    """


class NotInTimeWindow(SnmpError):
    """
    This exception is raised when a message is outside the time window

    .. seealso::
        `RFC-3414 - Section 3.2 - Processing an Incoming SNMP Message <https://tools.ietf.org/html/rfc3414#section-3.2>`_
            Reference description on how to handle incoming messages.
    """


class InvalidResponseId(SnmpError):
    """
    Exception which is raised when a response is received that did not
    correspond to the request-id
    """


class MissingPlugin(SnmpError):
    """
    Raised when a pluggable module could not be found

    :param ns: The plugin namespace
    :param needle: The identifier that was looked up
    :param haystack: The known identifiers
    """

    def __init__(self, ns: str, needle: Any, haystack: List[Any]) -> None:
        msg = (
            f"Namespace {ns!r} did not contain a plugin "
            f"with identifier {needle!r}. "
            f"Known identifiers: {sorted(haystack)!r}. "
            "See the 'puresnmp' documentation on plugins."
        )
        super().__init__(msg)
        self.ns = ns
        self.needle = repr(needle)
        self.haystack = haystack


class UnknownMessageProcessingModel(MissingPlugin):
    """
    Raised if a message was not formatted according to any known model
    """


class UnknownSecurityModel(MissingPlugin):
    """
    This exception is raised when something goes wrong with a security model
    """


class UnknownAuthModel(MissingPlugin):
    """
    Exception which is raised when working with an unsupported/unknown
    authentication
    """


class UnknownPrivacyModel(MissingPlugin):
    """
    Exception which is raised when the security model in use is unknown.
    """



"""
This module contains the definition for the community-based
message-processing model for SNMPv2
"""
from typing import Any, Awaitable, Callable, Dict, Optional, Union

from x690 import decode
from x690.types import Sequence

from puresnmp.credentials import V2C, Credentials
from puresnmp.pdu import PDU
from puresnmp.plugins.mpm import AbstractEncodingResult, MessageProcessingModel
from puresnmp.plugins.security import SecurityModel
from puresnmp.plugins.security import create as create_sm

IDENTIFIER = 1

TV2SecModel = SecurityModel[PDU, Sequence]


class V2CEncodingResult(AbstractEncodingResult):
    data: bytes
    security_model: Optional[TV2SecModel] = None


class V2CMPM(MessageProcessingModel[V2CEncodingResult, TV2SecModel]):
    """
    Message Processing Model for SNMP v2c
    """

    async def encode(
        self,
        request_id: int,
        credentials: Credentials,
        engine_id: bytes,
        context_name: bytes,
        pdu: PDU,
    ) -> V2CEncodingResult:
        if not isinstance(credentials, V2C):
            raise TypeError("SNMPv2c MPM should be used with V2C credentials!")

        security_model_id = 2
        if self.security_model is None:
            self.security_model = create_sm(security_model_id, self.lcd)

        packet = self.security_model.generate_request_message(
            pdu, b"", credentials
        )

        return V2CEncodingResult(bytes(packet))

    def decode(
        self,
        whole_msg: bytes,
        credentials: Credentials,
    ) -> PDU:
        """
        Convert an SNMPv2c message into a PDU instance
        """

        security_model_id = 2
        if self.security_model is None:
            self.security_model = create_sm(security_model_id)

        decoded, _ = decode(whole_msg, enforce_type=Sequence)

        msg = self.security_model.process_incoming_message(decoded, credentials)
        return msg


def create(
    transport_handler: Callable[[bytes], Awaitable[bytes]],
    lcd: Dict[str, Any],
) -> "MessageProcessingModel[V2CEncodingResult, Any]":
    """
    Creates a new instance of the V2C message-processing-model
    """
    return V2CMPM(transport_handler, lcd)



"""
This module implements the SNMPv3 "User Security Model" as defined in
:rfc:`3414`
"""
from dataclasses import dataclass, replace
from textwrap import indent
from typing import Any, Awaitable, Callable, Dict, Union, cast

from x690 import decode
from x690.types import Integer, Null, ObjectIdentifier, OctetString, Sequence
from x690.util import INDENT_STRING

import puresnmp.plugins.auth as auth
import puresnmp.plugins.priv as priv
from puresnmp.adt import (
    EncryptedMessage,
    HeaderData,
    Message,
    PlainMessage,
    ScopedPDU,
    V3Flags,
)
from puresnmp.credentials import V3, Credentials
from puresnmp.exc import NotInTimeWindow, SnmpError
from puresnmp.pdu import GetRequest, PDUContent
from puresnmp.plugins.security import SecurityModel
from puresnmp.transport import MESSAGE_MAX_SIZE
from puresnmp.util import get_request_id, localise_key, validate_response_id

IDENTIFIER = 3

MAX_ENGINE_TIME = 2147483647
"The value 2147483647 is a sentinel value defined in RFC-3414"


def reset_digest(message: Message) -> Message:
    """
    Replace the current message-digest in a message with zeroes.

    As the digest is embedded inside the message, it needs to be zeroed out
    when deriving the digest from a message. Otherwise the digest of the same
    message would change, because the digest changes.

    :param message: The message (with or without digest)
    :returns: A new message with zeroed digest
    """
    # As per https://tools.ietf.org/html/rfc3414#section-6.3.1,
    # the auth-key needs to be initialised to 12 zeroes
    secparams = USMSecurityParameters.decode(message.security_parameters)
    neutral = replace(secparams, auth_params=b"\x00" * 12)
    output = replace(
        message,
        security_parameters=bytes(neutral),
    )
    return output


class USMError(SnmpError):
    """
    Generic exception for errors cased by the USM module
    """


class UnsupportedSecurityLevel(USMError):
    """
    This error is raised when the data included in the credentials is invalid
    or incomplete.
    """


class EncryptionError(USMError):
    """
    This error is raised whenever something goes wrong during encryption
    """


class DecryptionError(USMError):
    """
    This error is raised whenever something goes wrong during decryption
    """


class AuthenticationError(USMError):
    """
    This error is raised whenever something goes wrong during authentication
    """


class UnknownUser(USMError):
    """
    This error is raised when a message is processed that is not consistent
    with the user-name passed in the credentials.
    """


@dataclass(frozen=True)
class DiscoData:
    """
    Helper class to wrap data received from a SNMPv3 discovery message.
    """

    #: The engine-id of the remote device
    authoritative_engine_id: bytes
    #: The number of boots of the remote device (encryption timeliness)
    authoritative_engine_boots: int
    #: The time-ticks since last boot of the remote device (encryption
    #: timeliness)
    authoritative_engine_time: int
    #: The number of requests received with an unknown client-engine-id
    unknown_engine_ids: int


@dataclass(frozen=True)
class USMSecurityParameters:
    """
    This class wraps the various values for the USM
    """

    authoritative_engine_id: bytes
    authoritative_engine_boots: int
    authoritative_engine_time: int
    user_name: bytes
    auth_params: bytes
    priv_params: bytes

    @staticmethod
    def decode(data: bytes) -> "USMSecurityParameters":
        """
        Construct a USMSecurityParameters instance from pure bytes
        """
        seq, _ = decode(data, enforce_type=Sequence)
        return USMSecurityParameters.from_snmp_type(seq)

    @staticmethod
    def from_snmp_type(seq: Sequence) -> "USMSecurityParameters":
        """
        Construct a USMSecurityParameters instance from an SNMP/X690 Sequence
        """
        return USMSecurityParameters(
            authoritative_engine_id=seq[0].pythonize(),
            authoritative_engine_boots=seq[1].pythonize(),
            authoritative_engine_time=seq[2].pythonize(),
            user_name=seq[3].pythonize(),
            auth_params=seq[4].pythonize(),
            priv_params=seq[5].pythonize(),
        )

    def __bytes__(self) -> bytes:
        return bytes(self.as_snmp_type())

    def as_snmp_type(self) -> Sequence:
        """
        Convert this instance into a plain SNMP (x690) object.
        """
        return Sequence(
            [
                OctetString(self.authoritative_engine_id),
                Integer(self.authoritative_engine_boots),
                Integer(self.authoritative_engine_time),
                OctetString(self.user_name),
                OctetString(self.auth_params),
                OctetString(self.priv_params),
            ]
        )

    def pretty(self, depth: int = 0) -> str:
        """
        Return a value for CLI display
        """
        idt = INDENT_STRING
        lines = ["Security Parameters"]
        lines.extend(
            [
                f"{idt}Engine ID   : {self.authoritative_engine_id!r}",
                f"{idt}Engine Boots: {self.authoritative_engine_boots}",
                f"{idt}Engine Time : {self.authoritative_engine_time}",
                f"{idt}Username    : {self.user_name!r}",
                f"{idt}Auth Params : {self.auth_params!r}",
                f"{idt}Priv Params : {self.priv_params!r}",
            ]
        )
        return indent("\n".join(lines), INDENT_STRING * depth)


def apply_encryption(
    message: PlainMessage,
    credentials: V3,
    security_name: bytes,
    security_engine_id: bytes,
    engine_boots: int,
    engine_time: int,
) -> Union[PlainMessage, EncryptedMessage]:
    """
    Derive a new encrypted message from a plain message given
    user-credentials and target-engine information.

    :param message: The unencrypted message
    :param credentials: The SNMPv3 credentials to be applied to the message
    :param security_name: The username used in the request
    :param security_engine_id: The engine-id of the remote device
    :param engine_boots: Encryption timeliness data as received from the
        discovery process
    :param engine_time: Encryption timeliness data as received from the
        discovery process
    :returns: Either an encrypted message, or a plain-message (depending on
        credential type)
    """

    # TODO: the security_name arg is redundant. Same value (different type)
    #       than credentials.username

    if credentials.priv is not None and not credentials.priv.method:
        raise UnsupportedSecurityLevel("Encryption method is missing")

    if credentials.priv is None:
        return replace(
            message,
            security_parameters=bytes(
                USMSecurityParameters(
                    security_engine_id,
                    engine_boots,
                    engine_time,
                    security_name,
                    b"",
                    b"",
                )
            ),
        )

    priv_method = priv.create(credentials.priv.method)
    localised_key = localise_key(credentials, security_engine_id)
    try:
        encrypted, salt = priv_method.encrypt_data(
            localised_key,
            security_engine_id,
            engine_boots,
            engine_time,
            bytes(message.scoped_pdu),
        )
        scoped_pdu = OctetString(encrypted)
    except Exception as exc:
        raise EncryptionError(f"Unable to encrypt message ({exc})") from exc

    unauthed_message = replace(
        message,
        scoped_pdu=scoped_pdu,
        security_parameters=bytes(
            USMSecurityParameters(
                security_engine_id,
                engine_boots,
                engine_time,
                security_name,
                b"",
                salt,
            )
        ),
    )
    return unauthed_message


def apply_authentication(
    unauthed_message: Union[PlainMessage, EncryptedMessage],
    credentials: V3,
    security_engine_id: bytes,
) -> Union[PlainMessage, EncryptedMessage]:
    """
    Add authentication-information to an SNMPv3 message

    :param unauthed_message: Either an encrypted or plain message that we
        want to update with authentication information.
    :param credentials: The user-credentials
    :param security_engine_id: The engine-id of the remote device
    :return: A message (of the same type) with authentication information
        added to and security_parameters of the message.
    """
    if credentials.auth is not None and not credentials.auth.method:
        raise UnsupportedSecurityLevel(
            "Incomplete data for authentication. "
            "Need both an auth-key and an auth-method!"
        )

    if credentials.auth is None:
        return unauthed_message

    auth_method = auth.create(credentials.auth.method)
    try:
        without_digest = reset_digest(unauthed_message)
        auth_result = auth_method.authenticate_outgoing_message(
            credentials.auth.key,
            bytes(without_digest),
            security_engine_id,
        )
        security_params = replace(
            USMSecurityParameters.decode(unauthed_message.security_parameters),
            auth_params=auth_result,
        )
        authed_message = replace(
            unauthed_message, security_parameters=bytes(security_params)
        )
        return authed_message
    except Exception as exc:
        raise AuthenticationError(
            f"Unable to authenticat the message ({exc})"
        ) from exc


def verify_authentication(
    message: Message, credentials: V3, security_params: USMSecurityParameters
) -> None:
    """
    Verify authenticity of the message using the credentials.

    :raises AuthenticationError: If the message is not authentic
    """

    if not message.header.flags.auth:
        return

    if not credentials.auth:
        raise UnsupportedSecurityLevel(
            "Message requires authentication but auth-method is missing!"
        )

    auth_method = auth.create(credentials.auth.method)
    without_digest = reset_digest(message)
    is_authentic = auth_method.authenticate_incoming_message(
        credentials.auth.key,
        bytes(without_digest),
        security_params.auth_params,
        security_params.authoritative_engine_id,
    )
    if not is_authentic:
        raise AuthenticationError(
            "Incoming message could not be authenticated!"
        )


def decrypt_message(
    message: Union[PlainMessage, EncryptedMessage], credentials: V3
) -> PlainMessage:
    """
    Decrypt a message using the given credentials
    """
    if isinstance(message, PlainMessage):
        return message

    if not credentials.priv:
        raise SnmpError("Attempting to decrypt a message without priv object")
    priv_method = priv.create(credentials.priv.method)
    key = credentials.priv.key
    if not isinstance(message.scoped_pdu, OctetString):
        raise SnmpError(
            "Unexpectedly received unencrypted PDU with a security level "
            "requesting encryption!"
        )
    security_parameters = USMSecurityParameters.decode(
        message.security_parameters
    )
    localised_key = localise_key(
        credentials, security_parameters.authoritative_engine_id
    )
    try:
        decrypted = priv_method.decrypt_data(
            localised_key,
            security_parameters.authoritative_engine_id,
            security_parameters.authoritative_engine_boots,
            security_parameters.authoritative_engine_time,
            security_parameters.priv_params,
            message.scoped_pdu.value,
        )
        message = cast(
            PlainMessage,
            replace(message, scoped_pdu=ScopedPDU.decode(decrypted)),
        )
    except Exception as exc:
        raise DecryptionError(f"Unable to decrypt message ({exc})") from exc
    return message


class UserSecurityModel(
    SecurityModel[PlainMessage, Union[PlainMessage, EncryptedMessage]]
):
    """
    Implementation of the use-security model as defined by
    :py:class:`puresnmp.security.SecurityModel`
    """

    def set_engine_timing(
        self,
        engine_id: bytes,
        engine_boots: int,
        engine_time: int,
    ) -> None:
        engine_config = self.local_config.setdefault(engine_id, {})
        engine_config["authoritative_engine_boots"] = engine_boots
        engine_config["authoritative_engine_time"] = engine_time
        if engine_time < engine_config.get("latest_received_engine_time", 0):
            raise NotInTimeWindow("packet timestamp predates the recorded state")
        engine_config["latest_received_engine_time"] = engine_time

    def generate_request_message(
        self,
        message: PlainMessage,
        security_engine_id: bytes,
        credentials: Credentials,
    ) -> Union[PlainMessage, EncryptedMessage]:
        if not isinstance(credentials, V3):
            raise TypeError(
                "Credentials must be a V3 instance for this scurity model!"
            )

        security_name = credentials.username.encode("ascii")
        engine_config = self.local_config[security_engine_id]
        engine_boots = engine_config["authoritative_engine_boots"]
        engine_time = engine_config["authoritative_engine_time"]

        encrypted_message = apply_encryption(
            message,
            credentials,
            security_name,
            security_engine_id,
            engine_boots,
            engine_time,
        )

        authed_message = apply_authentication(
            encrypted_message, credentials, security_engine_id
        )

        return authed_message

    def _verify_timesync(self, security_params: USMSecurityParameters) -> None:
        local_data = self.local_config.get(
            security_params.authoritative_engine_id
        )
        if local_data is None:
            raise SnmpError(
                "No local information for engine-id "
                f"{security_params.authoritative_engine_id.hex()} found!"
            )
        # fmt: off
        if (
            local_data["authoritative_engine_time"] == MAX_ENGINE_TIME or
            local_data["authoritative_engine_boots"] != security_params.authoritative_engine_boots or
            abs(local_data["authoritative_engine_time"] - security_params.authoritative_engine_time) > 150
        ):
        # fmt: on
            raise NotInTimeWindow("packet timestamp does not match the recorded local state")

    def process_incoming_message(
        self,
        message: Union[PlainMessage, EncryptedMessage],
        credentials: Credentials,
    ) -> PlainMessage:

        if not isinstance(credentials, V3):
            raise SnmpError("Supplied credentials is not a V3 instance!")

        security_params = USMSecurityParameters.decode(
            message.security_parameters
        )

        self._verify_timesync(security_params)
        security_name = security_params.user_name
        if security_name != credentials.username.encode("ascii"):
            # See https://tools.ietf.org/html/rfc3414#section-3.1
            raise UnknownUser(f"Unknown user {security_name!r}")

        verify_authentication(message, credentials, security_params)
        message = decrypt_message(message, credentials)
        validate_usm_message(message)
        return message

    async def _send_disco_message(
        self,
        user_name: bytes,
        engine_id: bytes,
        transport_handler: Callable[[bytes], Awaitable[bytes]],
    ) -> PlainMessage:
        """
        Build a discovery message, send it to the device and parse the response
        """
        request_id = get_request_id()
        security_params = USMSecurityParameters(
            authoritative_engine_id=engine_id,
            authoritative_engine_boots=0,
            authoritative_engine_time=0,
            user_name=user_name,
            auth_params=b"",
            priv_params=b"",
        )
        discovery_message = Message(
            Integer(3),
            HeaderData(
                request_id,
                MESSAGE_MAX_SIZE,
                V3Flags(user_name != b"", False, True),
                3,
            ),
            bytes(security_params),
            ScopedPDU(
                OctetString(),
                OctetString(),
                GetRequest(PDUContent(request_id, [])),
            ),
        )
        payload = bytes(discovery_message)
        raw_response = await transport_handler(payload)
        response, _ = decode(raw_response, enforce_type=Sequence)
        if isinstance(response, Null):
            raise SnmpError("Unexpectedly got a NULL object")

        # Arguably, this should pass through the message-processing model
        # allowing for more complex implementations. However this would require
        # a back-reference to the MPM (calling something "above" the current
        # abstraction level). As there is currently no implementation known to
        # me that requires this, I opted to go for a simpler architecture
        # instead. This means that discovery-messages cannot be encrypted.
        # Which they currently are not. So this should do.
        response_msg = PlainMessage.from_sequence(response)
        response_id = response_msg.header.message_id
        validate_response_id(request_id, response_id)
        return response_msg


    async def send_discovery_message(
        self,
        credentials: V3,
        transport_handler: Callable[[bytes], Awaitable[bytes]],
    ) -> DiscoData:
        # Via https://tools.ietf.org/html/rfc3414#section-4
        #
        # The User-based Security Model requires that a discovery process
        # obtains sufficient information about other SNMP engines in order to
        # communicate with them. Discovery requires an non-authoritative SNMP
        # engine to learn the authoritative SNMP engine's snmpEngineID value
        # before communication may proceed. This may be accomplished by
        # generating a Request message with a securityLevel of noAuthNoPriv, a
        # msgUserName of zero-length, a msgAuthoritativeEngineID value of zero
        # length, and the varBindList left empty. The response to this message
        # will be a Report message containing the snmpEngineID of the
        # authoritative SNMP engine as the value of the
        # msgAuthoritativeEngineID field within the msgSecurityParameters
        # field. It contains a Report PDU with the usmStatsUnknownEngineIDs
        # counter in the varBindList.
        response_msg = await self._send_disco_message(
            b"",
            b"",
            transport_handler
        )

        # The engine-id is available in two places: The response directly, and
        # also the Report PDU. In initial tests these values were identical,
        # and fetching them from the wrapping message would be easier. But
        # because the RFC explicitly states that it's the value from inside the
        # PDU I picked it out from there instead.
        security = USMSecurityParameters.decode(
            response_msg.security_parameters
        )
        wrapped_vars = response_msg.scoped_pdu.data.value.varbinds
        if not wrapped_vars:
            raise SnmpError("Invalid discovery response (no varbinds returned)")
        unknown_engine_id_var = wrapped_vars[0]
        if not unknown_engine_id_var.value:
            raise SnmpError("Discovery data did not contain valid data")
        unknown_engine_ids = unknown_engine_id_var.value.pythonize()

        out = DiscoData(
            authoritative_engine_id=security.authoritative_engine_id,
            authoritative_engine_boots=security.authoritative_engine_boots,
            authoritative_engine_time=security.authoritative_engine_time,
            unknown_engine_ids=unknown_engine_ids,
        )

        if credentials.username and (out.authoritative_engine_boots == 0 or out.authoritative_engine_time == 0):
            # According to RFC-3414, time must be synchronised for
            # authenticated requests. This is accomplished by sending an
            # additional discovery message which is both authenticated and has
            # the engine-id of the earlier discovery.
            # Most devices return the timing values in the initial discovery
            # response so we only send the second request if necessary (when
            # both values are 0)
            out = await self.initial_timesync(
                out,
                credentials.username,
                transport_handler
            )

        return out

    async def initial_timesync(
        self,
        disco: DiscoData,
        user_name: bytes,
        transport_handler: Callable[[bytes], Awaitable[bytes]]
    ) -> DiscoData:
        """
        Establish timing values from the authoritative engine.

        This step is required during the discovery process for authenticated
        requests.

        :param disco: An object containing discovery data of a remote device.
        :param user_name: The usernmae used during the information exchange
        :param transport_handler: A callable which is responsible to send data
            to the remote device.
        :return: A new copy of discovery data including the new timing
            information.
        """
        request_id = get_request_id()
        response = await self._send_disco_message(
            user_name,
            disco.authoritative_engine_id,
            transport_handler=transport_handler
        )
        output = replace(
            disco,
            authoritative_engine_boots=response.authoritative_engine_boots,
            authoritative_engine_time=response.authoritative_engine_time,
        )
        return output


def validate_usm_message(message: PlainMessage) -> None:
    """
    If the message contains known error-indicators, raise an appropriate
    exception.

    :raises SnmpError: If an error was found
    """
    pdu = message.scoped_pdu.data.value
    errors = {
        ObjectIdentifier(
            "1.3.6.1.6.3.15.1.1.1.0"
        ): "Unsupported security level",
        ObjectIdentifier("1.3.6.1.6.3.15.1.1.2.0"): "Not in time window",
        ObjectIdentifier("1.3.6.1.6.3.15.1.1.3.0"): "Unknown user-name",
        ObjectIdentifier("1.3.6.1.6.3.15.1.1.4.0"): "Unknown engine-id",
        ObjectIdentifier("1.3.6.1.6.3.15.1.1.5.0"): "Wrong message digest",
        ObjectIdentifier("1.3.6.1.6.3.15.1.1.6.0"): "Unable to decrypt",
    }
    for varbind in pdu.varbinds:
        if varbind.oid in errors:
            msg = errors[varbind.oid]
            raise SnmpError(f"Error response from remote device: {msg}")


def create(local_config: Dict[bytes, Dict[str, Any]]) -> UserSecurityModel:
    """
    Creates a new instance of the USM
    """
    return UserSecurityModel(local_config)



"""
Collection of utility functions for the puresnmp package.
"""
import asyncio
import hashlib
import pkgutil
from dataclasses import dataclass
from functools import lru_cache
from time import time
from types import ModuleType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from x690.types import ObjectIdentifier

from puresnmp.credentials import V3
from puresnmp.exc import InvalidResponseId, SnmpError
from puresnmp.typevars import TAnyIp
from puresnmp.varbind import VarBind

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

T = TypeVar("T", bound=Any)
TTableRow = TypeVar("TTableRow", bound=Any)


class TDigestable(Protocol):
    """
    Typing protocol copied from https://github.com/python/typeshed/blob/569fcea637023b3df2fe45c6f08c26037bfa6a74/stdlib/hashlib.pyi#L5
    """

    digest_size: int
    block_size: int

    # [Python documentation note] Changed in version 3.4: The name attribute has
    # been present in CPython since its inception, but until Python 3.4 was not
    # formally specified, so may not exist on some platforms
    name: str

    def __init__(self, data: bytes = ...) -> None:
        ...

    def copy(self) -> "TDigestable":
        ...

    def digest(self) -> bytes:
        ...

    def hexdigest(self) -> str:
        ...

    def update(self, __data: bytes) -> None:
        ...


@dataclass(frozen=True)
class WalkRow:
    """
    A wrapper around an SNMP Walk item.

    This also keeps track whether this walk result should be considered the
    last row or not.
    """

    #: The value of the "current" row in a walk operation
    value: Any
    #: Whether there are still values following this row
    unfinished: bool


@dataclass(frozen=True)
class BulkResult:
    """
    A representation for results of a "bulk" request.

    These requests get both "non-repeating values" (scalars) and "repeating
    values" (lists). This wrapper makes these terms a bit friendlier to use.
    """

    #: A mapping from object-identifiers to scalar (single) values for those
    #: OIDs
    scalars: Dict[ObjectIdentifier, Any]
    #: A mapping from object-identifiers to "walk" results below those OIDs
    listing: Dict[ObjectIdentifier, Any]


def group_varbinds(
    varbinds: List[VarBind],
    effective_roots: List[ObjectIdentifier],
    user_roots: Optional[List[ObjectIdentifier]] = None,
) -> Dict[ObjectIdentifier, List[VarBind]]:
    """
    Takes a list of varbinds and a list of base OIDs and returns a mapping from
    those base IDs to lists of varbinds.

    Varbinds returned from a walk operation which targets multiple OIDs is
    returned as "interleaved" list. This functions extracts these interleaved
    items into a more usable dictionary.

    >>> from x690.types import Integer
    >>> result = group_varbinds(
    ...     [
    ...         VarBind(ObjectIdentifier("1.1.1"), Integer(1)),
    ...         VarBind(ObjectIdentifier("2.2.2"), Integer(1)),
    ...         VarBind(ObjectIdentifier("1.1.2"), Integer(1)),
    ...         VarBind(ObjectIdentifier("2.2.3"), Integer(1)),
    ...     ],
    ...     [
    ...         ObjectIdentifier("1.1"),
    ...         ObjectIdentifier("2.2"),
    ...     ],
    ... )
    >>> sorted(result.keys())
    [ObjectIdentifier('1.1'), ObjectIdentifier('2.2')]
    >>> result[ObjectIdentifier("1.1")]
    [VarBind(oid=ObjectIdentifier('1.1.1'), value=Integer(1)), VarBind(oid=ObjectIdentifier('1.1.2'), value=Integer(1))]
    >>> result[ObjectIdentifier("2.2")]
    [VarBind(oid=ObjectIdentifier('2.2.2'), value=Integer(1)), VarBind(oid=ObjectIdentifier('2.2.3'), value=Integer(1))]


    :param varbinds: A list of VarBind instances.
    :param effective_roots: The list of OIDs that were requested from the SNMP
        agent.
    :param user_roots: The set of VarBind instances that were requested by the
        user. This is used internally for walk requests. On each request
        following the first, the requested OIDs will differ from the OIDs
        requested by the user. This list will keep track of the original OIDs
        to determine when the walk needs to terminate.
    """
    user_roots = user_roots or []
    n = len(effective_roots)

    results = {}
    for i in range(n):
        results[effective_roots[i]] = varbinds[i::n]

    if user_roots:
        new_results = {}
        for key, value in results.items():
            containment = [base for base in user_roots if key in base]
            if len(containment) > 1:
                raise RuntimeError(
                    "Unexpected OID result. A value was "
                    "contained in more than one base than "
                    "should be possible!"
                )
            if not containment:
                continue
            new_results[containment[0]] = value
            results = new_results

    return results


def get_unfinished_walk_oids(
    grouped_oids: Dict[ObjectIdentifier, List[VarBind]]
) -> List[Tuple[ObjectIdentifier, WalkRow]]:
    """
    Create a list of OIDs which still have subsequent values in a "walk"
    operation

    :param grouped_oids: A dictionary containing VarBinds as values. The keys
        are the base OID of those VarBinds as requested by the user. We need to
        keep track of the base to be able to tell when a walk over OIDs is
        finished (that is, when we hit the first OID outside the base).
    """

    # grouped_oids contains a list of values for each requested OID. We need to
    # determine if we need to continue fetching: Inspect the last item of each
    # list if those OIDs are still children of the requested IDs we need to
    # continue fetching using *those* IDs (as we're using GetNext behaviour).
    # If they are *not* children of the requested OIDs, we went too far (in the
    # case of a bulk operation) and need to remove all outliers.
    #
    # The above behaviour is the same for both bulk and simple operations. For
    # simple operations we simply have a list of 1 element per OID, but the
    # behaviour is identical

    # Build a mapping from the originally requested OID to the last fetched OID
    # from that tree.
    last_received_oids = {
        k: WalkRow(v[-1], v[-1].oid in k) for k, v in grouped_oids.items() if v
    }

    output = [
        item
        for item in sorted(last_received_oids.items())
        if item[1].unfinished
    ]
    return output


def tablify(
    varbinds: Iterable[VarBind],
    num_base_nodes: int = 0,
    base_oid: str = "",
    _rowtype: Type[TTableRow] = Dict[str, Any],  # type: ignore
) -> List[TTableRow]:
    # pylint: disable=line-too-long
    """
    Converts a list of varbinds into a table-like structure. *num_base_nodes*
    can be used for table which row-ids consist of multiple OID tree nodes. By
    default, the last node is considered the row ID, and the second-last is the
    column ID. Example:

    By default, for the table-cell at OID ``1.2.3.4.5``, ``4`` is the column
    index and ``5`` is the row index.

    Using ``num_base_nodes=2`` will only use the first two nodes (``1.2``) as
    table-identifier, so ``3`` becomes the column index, and ``4.5`` becomes
    the row index.

    The output should *not* be considered ordered in any way. If you need it
    sorted, you must sort it after retrieving the table from this function!

    Each element of the output is a dictionary where each key is the column
    index. By default the index ``0`` will be added, representing the row ID.

    Example::

        >>> data = [
        ...     (ObjectIdentifier('1.2.1.1'), 'row 1 col 1'),
        ...     (ObjectIdentifier('1.2.1.2'), 'row 2 col 1'),
        ...     (ObjectIdentifier('1.2.2.1'), 'row 1 col 2'),
        ...     (ObjectIdentifier('1.2.2.2'), 'row 2 col 2'),
        ... ]
        >>> tablify(data)
        [{'0': '1', '1': 'row 1 col 1', '2': 'row 1 col 2'}, {'0': '2', '1': 'row 2 col 1', '2': 'row 2 col 2'}]


    Example with longer row ids (using the *first* two as table identifiers)::

        >>> data = [
        ...     (ObjectIdentifier('1.2.1.5.10'), 'row 5.10 col 1'),
        ...     (ObjectIdentifier('1.2.1.6.10'), 'row 6.10 col 1'),
        ...     (ObjectIdentifier('1.2.2.5.10'), 'row 5.10 col 2'),
        ...     (ObjectIdentifier('1.2.2.6.10'), 'row 6.10 col 2'),
        ... ]
        >>> tablify(data, num_base_nodes=2)
        [{'0': '5.10', '1': 'row 5.10 col 1', '2': 'row 5.10 col 2'}, {'0': '6.10', '1': 'row 6.10 col 1', '2': 'row 6.10 col 2'}]
    """
    # pylint: enable=line-too-long

    if isinstance(base_oid, str) and base_oid:

        base_oid_parsed = ObjectIdentifier(base_oid)
        # Each table has a sub-index for the table "entry" so the number of
        # base-nodes needs to be incremented by 1
        num_base_nodes = len(base_oid_parsed)

    rows: Dict[str, TTableRow] = {}
    for oid, value in varbinds:
        if num_base_nodes:
            tail = oid.nodes[num_base_nodes:]
            col_id_nodes, row_id_nodes = tail[0], tail[1:]
            col_id = str(col_id_nodes)
            row_id = ".".join([str(node) for node in row_id_nodes])
        else:
            col_id = str(oid.nodes[-2])
            row_id = str(oid.nodes[-1])
        tmp: TTableRow = {  # type: ignore
            "0": row_id,
        }
        row = rows.setdefault(row_id, tmp)
        row[str(col_id)] = value
    return list(rows.values())


def password_to_key(
    hash_implementation: Callable[[bytes], TDigestable], padding_length: int
) -> Callable[[bytes, bytes], bytes]:
    """
    Create a helper function to convert passwords to SNMP compliant keys
    according to :rfc:`3414`.

    >>> hasher = password_to_key(hashlib.sha1, 20)
    >>> key = hasher(b"mypasswd", b"target-engine-id")
    >>> key.hex()
    '999ec23ca66b9d3f187ab5208840c30b0450b452'

    :param hash_implementation: A callable that creates an object with a
        ".digest()" method from a bytes-object. Usable examples are
        `hashlib.md5` and `hashlib.sha1`
    :param padding_length: The padding length to be used during hashing (as
        defined in the SNMP rfc)
    :returns: A callable which can be used to derive an SNMP compliant key
        from a password.
    """

    @lru_cache(maxsize=None)
    def hasher(password: bytes, engine_id: bytes) -> bytes:
        """
        Derive a key from a password and engine-id.

        :param password: The user password
        :param engine_id: The target engine ID
        :returns: The derived key
        """
        # Repeat the password for a total of 1MB worth of data (as per SNMP rfc)
        hash_size = 1024 * 1024
        num_words = hash_size // len(password)
        tmp = (password * (num_words + 1))[:hash_size]
        hash_instance = hash_implementation(tmp)
        key = hash_instance.digest()
        localised_buffer = (
            key[:padding_length] + engine_id + key[:padding_length]
        )
        final_key = hash_implementation(localised_buffer).digest()
        return final_key

    hasher.__name__ = f"<hasher:{hash_implementation}>"  # type: ignore
    return hasher


def generate_engine_id_ip(pen: int, ip: TAnyIp) -> bytes:
    """
    Generates a valid SNMP Engine ID using a private enterprise number and an
    ip-address.

    >>> from ipaddress import ip_address
    >>> generate_engine_id_ip(696, ip_address("192.0.2.1"))
    b'\\x80\\x00\\x02\\xb8\\x01\\xc0\\x00\\x02\\x01'

    .. seealso::
        `Engine ID structure <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            ASN.1 definition for engine-id encoding
        `Engine ID types <https://tools.ietf.org/html/rfc5343#section-4>`_
            List of valid engine-id variants
        `PEN list <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            List of publicly registered private enterprise numbers
    """
    buffer = bytearray(pen.to_bytes(4, "big"))
    # Setting the first bit to 1 is the same as setting the first byte to 16*8
    buffer[0] = 16 * 8
    fmt = 1 if ip.version == 4 else 2
    buffer.append(fmt)
    buffer.extend(ip.packed)
    return bytes(buffer)


def generate_engine_id_mac(pen: int, mac_address: str) -> bytes:
    """
    Generates a valid SNMP Engine ID using a private enterprise number and a
    mac-address.

    >>> generate_engine_id_mac(696, "01:02:03:04:05:06")
    b'\\x80\\x00\\x02\\xb8\\x03\\x01\\x02\\x03\\x04\\x05\\x06'

    .. seealso::
        `Engine ID structure <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            ASN.1 definition for engine-id encoding
        `Engine ID types <https://tools.ietf.org/html/rfc5343#section-4>`_
            List of valid engine-id variants
        `PEN list <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            List of publicly registered private enterprise numbers
    """
    buffer = bytearray(pen.to_bytes(4, "big"))

    # Setting the first bit to 1 is the same as setting the first byte to 16*8
    buffer[0] = 16 * 8

    if "-" in mac_address:
        octets = [int(oct, 16) for oct in mac_address.split("-")]
    else:
        octets = [int(oct, 16) for oct in mac_address.split(":")]

    buffer.append(3)
    buffer.extend(octets)
    return bytes(buffer)


def generate_engine_id_text(pen: int, text: str) -> bytes:
    """
    Generates a valid SNMP Engine ID using a private enterprise number and a
    custom text (no longer than 27 characters).

    >>> generate_engine_id_text(696, "hello")
    b'\\x80\\x00\\x02\\xb8\\x04hello'

    .. seealso::
        `Engine ID structure <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            ASN.1 definition for engine-id encoding
        `Engine ID types <https://tools.ietf.org/html/rfc5343#section-4>`_
            List of valid engine-id variants
        `PEN list <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            List of publicly registered private enterprise numbers
    """
    if len(text) > 27:
        raise SnmpError(
            "Invalid engine ID. Text must have fewer than 27 characters"
        )
    buffer = bytearray(pen.to_bytes(4, "big"))
    # Setting the first bit to 1 is the same as setting the first byte to 16*8
    buffer[0] = 16 * 8
    buffer.append(4)
    buffer.extend(text.encode("ascii"))
    return bytes(buffer)


def generate_engine_id_octets(pen: int, octets: bytes) -> bytes:
    """
    Generates a valid SNMP Engine ID using a private enterprise number and a
    custom byte-string (no longer than 27 bytes)

    >>> generate_engine_id_octets(696, b"hello")
    b'\\x80\\x00\\x02\\xb8\\x05hello'

    .. seealso::
        `Engine ID structure <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            ASN.1 definition for engine-id encoding
        `Engine ID types <https://tools.ietf.org/html/rfc5343#section-4>`_
            List of valid engine-id variants
        `PEN list <https://www.iana.org/assignments/enterprise-numbers/enterprise-numbers>`_
            List of publicly registered private enterprise numbers
    """
    if len(octets) > 27:
        raise SnmpError(
            f"Invalid engine ID. The value {octets!r} is longer than 27 octets"
        )
    buffer = bytearray(pen.to_bytes(4, "big"))
    # Setting the first bit to 1 is the same as setting the first byte to 16*8
    buffer[0] = 16 * 8
    buffer.append(5)
    buffer.extend(octets)
    return bytes(buffer)


def validate_response_id(request_id: int, response_id: int) -> None:
    """
    Compare request and response IDs and raise an appropriate error.

    Raises an appropriate error if the IDs differ. Otherwise returns

    This helper method ensures we're always returning the same exception type
    on invalid response IDs.
    """
    if response_id != request_id:
        raise InvalidResponseId(
            f"Invalid response ID {response_id} for request id {request_id}"
        )


def iter_namespace(
    ns_pkg: ModuleType,
) -> Generator[pkgutil.ModuleInfo, None, None]:
    """
    Iterates over modules inside the given namespace
    """
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")  # type: ignore


def localise_key(credentials: V3, engine_id: bytes) -> bytes:
    """
    Derive a localised key from the user-credentials.

    This follows the logic as dictated by :rfc:`3414` melding the recipient
    engine-id with the user-credentials into a kay used for de-/en-cryption
    of packets.

    .. seealso: https://tools.ietf.org/html/rfc3414#section-2.6

    >>> from puresnmp.credentials import V3, Auth, Priv
    >>> engine_id = b"\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x02"
    >>> localised = localise_key(
    ...     V3(b"user", Auth(b"maplesyrup", "md5"), Priv(b"privkey", "des")),
    ...     engine_id
    ... )
    >>> localised.hex()
    '4924c679907476d038b258097995a15c'
    """
    if credentials.priv is None:
        raise SnmpError(
            "Attempting to derive a localised key from an empty "
            "privacy object!"
        )
    if credentials.auth is None:
        raise SnmpError(
            "Attempting to derive a localised key from an empty " "auth object!"
        )
    if credentials.auth.method == "md5":
        hasher = password_to_key(
            cast(Callable[[bytes], TDigestable], hashlib.md5), 16
        )
    elif credentials.auth.method == "sha1":
        hasher = password_to_key(
            cast(Callable[[bytes], TDigestable], hashlib.sha1), 20
        )
    else:
        raise SnmpError(
            "Unknown authentication method: %r" % credentials.auth.method
        )

    output = hasher(credentials.priv.key, engine_id)
    return output


def get_request_id() -> int:  # pragma: no cover
    """
    Generates a SNMP request ID.

    This returns a simple integer used to validate if a given response
    matches with the given request.
    """
    return int(time())


def sync(coro: Awaitable[T]) -> T:
    """
    Execute an asyncio corouting in the current event-loop and return the result

    >>> async def foo():
    ...     return "hello"
    >>> sync(foo())
    'hello'

    .. warning::

        This is intended for debugging and testing. If possible you should
        use your own async environment.
    """
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(coro)
    return result



"""
This module contains the implementation for the SNMPv3 message-processing model
"""
from typing import Any, Awaitable, Callable, Dict, Optional, Union

from x690.types import Integer, OctetString

from puresnmp.adt import (
    EncryptedMessage,
    HeaderData,
    Message,
    PlainMessage,
    ScopedPDU,
    V3Flags,
)
from puresnmp.credentials import V3, Credentials
from puresnmp.pdu import PDU, GetRequest
from puresnmp.plugins.mpm import AbstractEncodingResult, MessageProcessingModel
from puresnmp.plugins.security import SecurityModel
from puresnmp.plugins.security import create as create_sm
from puresnmp.transport import MESSAGE_MAX_SIZE

IDENTIFIER = 3


def is_confirmed(pdu: PDU) -> bool:
    """
    Return True if the given PDU instance expects a response.
    """
    # XXX TODO This might be doable cleaner with subclassing in puresnmp.pdu
    return isinstance(pdu, GetRequest)


TV3SecModel = SecurityModel[PlainMessage, Union[PlainMessage, EncryptedMessage]]


class V3EncodingResult(AbstractEncodingResult):
    """
    A simple data-structure representing the output for encoded messages
    """

    data: bytes
    security_model: Optional[TV3SecModel] = None


class V3MPM(MessageProcessingModel[V3EncodingResult, TV3SecModel]):
    """
    This class contains the concrete implementation for the v3
    message-processing-model.
    """

    def decode(
        self,
        whole_msg: bytes,  # as received from the network
        credentials: Credentials,
    ) -> PDU:
        security_model_id = 3
        if self.security_model is None:
            self.security_model = create_sm(security_model_id, self.lcd)
        message = Message.decode(whole_msg)
        msg = self.security_model.process_incoming_message(message, credentials)
        return msg.scoped_pdu.data

    async def encode(
        self,
        request_id: int,
        credentials: Credentials,
        engine_id: bytes,
        context_name: bytes,
        pdu: PDU,
    ) -> V3EncodingResult:

        if not isinstance(credentials, V3):
            raise TypeError("Credentials for SNMPv3 must be V3 instances!")

        security_model_id = 3
        if self.security_model is None:
            self.security_model = create_sm(security_model_id, self.lcd)

        # We need to determine some values from the remote host for security.
        # These can be retrieved by sending a so called discovery message.
        if not self.disco:
            self.disco = await self.security_model.send_discovery_message(
                credentials, self.transport_handler
            )
            self.security_model.set_engine_timing(
                self.disco.authoritative_engine_id,
                self.disco.authoritative_engine_boots,
                self.disco.authoritative_engine_time,
            )

        security_engine_id = self.disco.authoritative_engine_id

        if engine_id == b"":
            engine_id = security_engine_id

        scoped_pdu = ScopedPDU(
            OctetString(engine_id), OctetString(context_name), pdu
        )
        flags = V3Flags(
            auth=credentials.auth is not None,
            priv=credentials.priv is not None,
            reportable=is_confirmed(pdu),
        )
        header = HeaderData(
            request_id,
            MESSAGE_MAX_SIZE,
            flags,
            security_model_id,
        )

        snmp_version = 3
        msg = PlainMessage(Integer(snmp_version), header, b"", scoped_pdu)
        output = self.security_model.generate_request_message(
            msg,
            security_engine_id,
            credentials,
        )

        outgoing_message = bytes(output)
        return V3EncodingResult(outgoing_message, self.security_model)


def create(
    transport_handler: Callable[[bytes], Awaitable[bytes]],
    lcd: Dict[str, Any],
) -> "MessageProcessingModel[V3EncodingResult, TV3SecModel]":
    """
    Creates a new instance of the V3 message-processing-model
    """
    return V3MPM(transport_handler, lcd)



"""
Low-Level network transport for asyncio.

This module mainly exist to enable a "seam" for mocking/patching out during
testing.

The module is excluded from coverage. It contains all the "dirty" stuff that's
hard to test.
"""

import asyncio
import logging
import socket
from asyncio.events import AbstractEventLoop
from asyncio.transports import BaseTransport
from typing import Any, Callable, NamedTuple, Optional, Tuple, Union

from x690.util import visible_octets

from .const import DEFAULT_LISTEN_ADDRESS
from .exc import Timeout
from .typevars import SocketInfo, SocketResponse, TAnyIp

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

LOG = logging.getLogger(__name__)
MESSAGE_MAX_SIZE = 65507


class Endpoint(NamedTuple):
    """
    A tuple representing an UDP endpoint where a connection should be made to.
    """

    ip: TAnyIp
    port: int


class TSender(Protocol):
    """
    A typing-protocol for callables which send data out to the network
    """

    # pylint: disable=too-few-public-methods

    async def __call__(
        self,
        endpoint: Endpoint,
        packet: bytes,
        timeout: int = 1,
        loop: Optional[AbstractEventLoop] = None,
        retries: int = 10,
    ) -> bytes:  # pragma: no cover
        ...


class SNMPTrapReceiverProtocol(asyncio.DatagramProtocol):
    """
    A protocol to handle incoming SNMP traps.

    The protocol requires a callable which is called with a
    :py:class:`~.SocketResponse` instance whenever a trap is received.
    """

    def __init__(self, callback: Callable[[SocketResponse], Any]) -> None:
        super().__init__()
        self.callback = callback
        self.transport: Optional[BaseTransport] = None

    def connection_made(self, transport: BaseTransport) -> None:
        self.transport = transport

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        if LOG.isEnabledFor(logging.DEBUG):
            hexdump = visible_octets(data)
            LOG.debug("Received packet:\n%s", hexdump)
        self.callback(SocketResponse(data, SocketInfo(addr[0], addr[1])))


class SNMPClientProtocol(asyncio.DatagramProtocol):
    """
    An SNMP Client Protocol suitable for use with
    :py:meth:`asyncio.AbstractEventLoop.create_datagram_endpoint` that
    provides a method to convert the callback based API into a coroutine
    based API.
    """

    def __init__(self, packet: bytes) -> None:
        loop = asyncio.get_running_loop()
        self.packet = packet
        self.transport = None  # type: Optional[asyncio.DatagramTransport]
        self.loop = loop
        self.future = loop.create_future()

    def connection_made(self, transport):  # type: ignore
        # type: (asyncio.DatagramTransport) -> None
        """
        Sends the SNMP request packet when a connection is made.
        """
        self.transport = transport

        if LOG.isEnabledFor(logging.DEBUG):
            hexdump = visible_octets(self.packet)
            ip, port = self.transport.get_extra_info("peername", ("", ""))
            LOG.debug("Sending packet to %s:%s\n%s", ip, port, hexdump)

        self.transport.sendto(self.packet)

    def connection_lost(self, exc):
        # type: (Optional[Exception]) -> None
        """
        Handles the socket being closed optionally passing on an exception.
        """
        if LOG.isEnabledFor(logging.DEBUG):
            if exc is None:
                LOG.debug("Socket closed")
            else:
                LOG.debug("Connection lost: %s", exc)

        if exc is not None:
            self.future.set_exception(exc)

    def datagram_received(self, data, addr):
        # type: (Union[bytes, str], Tuple[str, int]) -> None
        """
        Receive the data and close the connection.
        """
        if LOG.isEnabledFor(logging.DEBUG) and isinstance(data, bytes):
            hexdump = visible_octets(data)
            LOG.debug(
                "Received packet from %s:%d:\n%s", addr[0], addr[1], hexdump
            )

        self.future.set_result(data)
        if self.transport:
            self.transport.close()

    def error_received(self, exc):
        # type: (Exception) -> None
        """
        Pass the exception along if there is an error.
        """
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("Error received!", exc_info=exc)

        self.future.set_exception(exc)

    async def get_data(self, timeout):
        # type: (int) -> bytes
        """
        Retrieve the response data back into the calling coroutine.
        """
        try:
            return await asyncio.wait_for(self.future, timeout)
        except (asyncio.TimeoutError, socket.timeout) as exc:
            if self.transport:
                self.transport.abort()
            raise Timeout(
                f"{timeout} second timeout exceeded on UDP transport."
            ) from exc


async def send_udp(
    endpoint: Endpoint,
    packet: bytes,
    timeout: int = 1,
    loop: Optional[AbstractEventLoop] = None,
    retries: int = 10,
) -> bytes:  # pragma: no cover
    # pylint: disable=arguments-differ
    """
    A coroutine that opens a UDP socket to *ip:port*, sends a packet with
    *bytes* and returns the raw bytes as returned from the remote host.

    If the connection fails due to a timeout, a Timeout exception is
    raised.
    """
    if loop is None:
        loop = asyncio.get_event_loop()

    while retries > 0:
        _, protocol = await loop.create_datagram_endpoint(
            lambda: SNMPClientProtocol(packet),
            remote_addr=(str(endpoint.ip), endpoint.port),
        )
        try:
            response = await protocol.get_data(timeout)
            break
        except Timeout:
            if retries == 1:
                raise
            retries -= 1
            LOG.debug("Resending UDP packet. %d retries left", retries)

    return response


def default_trap_handler(response: SocketResponse) -> None:
    """
    A no-op implementation for trap handlers which only logs traps.
    """
    LOG.debug("Trap Received from %s: %r", response.info, response)


async def listen(
    bind_address: str = DEFAULT_LISTEN_ADDRESS,
    port: int = 162,
    callback: Callable[[SocketResponse], Any] = default_trap_handler,
    loop: Optional[AbstractEventLoop] = None,
) -> None:  # pragma: no cover
    """
    Sets up a listening UDP socket and returns a generator over recevied
    packets::

        >>> transport = Transport()  # doctest: +SKIP
        >>> for seq, packet in enumerate(transport.listen()):  # doctest: +SKIP
        ...     print(seq, repr(packet))
        0, b'...'
        1, b'...'
        2, b'...'

    .. note::

        This defaults to the standard SNMP Trap port 162. This is a
        privileged port so processes using this port must run as root!
    """
    if loop is None:
        loop = asyncio.get_event_loop()
    await loop.create_datagram_endpoint(
        lambda: SNMPTrapReceiverProtocol(callback),
        local_addr=(bind_address, port),
    )



import importlib
import logging
from threading import Lock
from types import ModuleType
from typing import Any, Callable, Dict

from puresnmp.util import iter_namespace

DISCOVERY_LOCK = Lock()
LOG = logging.getLogger(__name__)


def discover_plugins(
    namespace: str, is_valid_module: Callable[[ModuleType], bool]
) -> Dict[Any, ModuleType]:
    """
    Load all plugins in the given namespace (only if it is a valid module)

    Module validity check is handed off to the *is_valid_module* callable.
    Any module that is not valid is silently skipped.
    """
    with DISCOVERY_LOCK:
        output: Dict[Any, ModuleType] = {}
        try:
            nsmod = importlib.import_module(namespace)
        except ImportError as exc:
            LOG.debug(
                "Unable to load plugins in %r: %s",
                namespace,
                exc,
                exc_info=True,
            )
            return {}

        for _, name, _ in iter_namespace(nsmod):
            try:
                mod = importlib.import_module(name)
            except ImportError as exc:
                LOG.debug(
                    "Unable to load plugin %r: %s", name, exc, exc_info=True
                )
                continue
            if not is_valid_module(mod):
                continue
            if mod.IDENTIFIER in output:
                raise ImportError(
                    "Plugin %r causes a name-clash with the identifier %r. "
                    "This is already used by %r"
                    % (mod, mod.IDENTIFIER, output[mod.IDENTIFIER])
                )
            output[mod.IDENTIFIER] = mod

        return output


class Loader:
    def __init__(
        self, namespace: str, validator: Callable[[ModuleType], bool]
    ) -> None:
        self.namespace = namespace
        self.validator = validator
        self.discovered_plugins: Dict[Any, ModuleType] = {}

    def create(self, name: Any) -> Any:
        """
        Return an instance of the given privacy module by identifier.

        This looks up the module by "IDENTIFIER" as specified in the given plugin.

        If no plugin with the given identifier is found, a *KeyError* is raised
        """

        if not self.discovered_plugins:
            self.discovered_plugins = discover_plugins(
                self.namespace, self.validator
            )
        if name not in self.discovered_plugins:
            return None
        return self.discovered_plugins[name]



"""
This module provides a plugin architecture for SNMP message-processing models
(mpm).

Each mpm plugin can be distributed as separate package by providing
modules inside the namespace-package "puresnmp.mpm". Note that in order
to be a valid namespace-package, such a package *must not* have a
``__init__.py`` file!

Example folder-structure for a mpm plugin:

.. code-block:: text

    my-security-plugin/
     +- setup.py (or pyproject.toml)
     +- puresnmp/
         +- mpm/
             +- mymodule.py
             +- myothermodule.py

Note that there is no ``__init__.py`` file!

In order for modules to be detected as plugin, they must follow the following
rules:

* Have a function ``create`` returning an instance of a class implementing the
  message-processing model. The function must take a network-transport handler
  and a "local configuration directory" as arguments. A reference
  implementation can be found in :py:class:`puresnmp.mpm.v3.V3MPM`.
* Contain a int-variable ``IDENTIFIER``. This variable should map to the
  message-processing model identifiers as defined in the SNMPv3 standard. See
  https://www.iana.org/assignments/snmp-number-spaces/snmp-number-spaces.xhtml
  and :rfc:`3411`
"""
from types import ModuleType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    NamedTuple,
    Optional,
    TypeVar,
)

from puresnmp.credentials import Credentials
from puresnmp.exc import UnknownMessageProcessingModel
from puresnmp.pdu import PDU
from puresnmp.plugins.pluginbase import Loader
from puresnmp.plugins.security import SecurityModel

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


class AbstractEncodingResult(NamedTuple):
    """
    Generic type for the output of the conversion from messages to bytes by
    the message processing modules.
    """

    #: The bytes that are to be sent out to the network
    data: bytes

    #: The security model that was applied by the message-processing module
    security_model: Optional[SecurityModel[Any, Any]] = None


TEncodeResult = TypeVar("TEncodeResult", bound=AbstractEncodingResult)
TSecurityModel = TypeVar("TSecurityModel", bound=SecurityModel[Any, Any])


class TMPMPlugin(Protocol):
    """
    Protocol for the plugin factory function.
    """

    # pylint: disable=too-few-public-methods

    def create(
        self,
        transport_handler: Callable[[bytes], Awaitable[bytes]],
        lcd: Dict[str, Any],
    ) -> "MessageProcessingModel[TEncodeResult, TSecurityModel]":  # pragma: no cover
        """
        See :py:func:`~.create`
        """
        ...


#: Global registry of detected plugins
DISCOVERED_PLUGINS: Dict[int, TMPMPlugin] = {}


class MessageProcessingModel(Generic[TEncodeResult, TSecurityModel]):
    """
    Each Message Processing Model defines the format of a particular version of
    an SNMP message and coordinates the preparation and extraction of each such
    version-specific message format.

    See https://tools.ietf.org/html/rfc3411#section-6.2
    """

    #: The IANA registered identifier for thie mosule
    IDENTIFIER: int

    #: Discovery data (if required by the message-processing-module)
    disco: Optional[Any]

    #: The security model applied by this MPM (if any)
    security_model: Optional[TSecurityModel]

    def __init__(
        self,
        transport_handler: Callable[[bytes], Awaitable[bytes]],
        lcd: Dict[str, Any],
    ) -> None:
        self.transport_handler = transport_handler
        self.lcd = lcd
        self.disco = None
        self.security_model = None

    async def encode(
        self,
        request_id: int,
        credentials: Credentials,
        engine_id: bytes,
        context_name: bytes,
        pdu: PDU,
    ) -> TEncodeResult:  # pragma: no cover
        """
        Convert an SNMP PDU into raw bytes for the network.

        :param request_id: A unique ID for the request.
        :param credentials: Credentials which are used by the model if needed
            (encryption, authentication, ...).
        :param engine_id: The engine-id for use in SNMPv3.
            :py:mod:`puresnmp.util` contains helper methods to create those
            according to the RFC.
        :param context_name: An identifier for the SNMPv3 context
        :param pdu: The plain SNMP PDU to process
        :returns: Raw bytes that can be sent to the network
        """
        raise NotImplementedError(
            "encode is not yet implemented in %r" % type(self)
        )

    def decode(
        self,
        whole_msg: bytes,
        credentials: Credentials,
    ) -> PDU:  # pragma: no cover
        """
        Convert bytes (as received raw from the network) into an SNMP PDU

        :param whole_msg: The message as received from the network
        :param credentials: Credentials which will be used if required by the
            model.
        :returns: A plain SNMP PDU
        """
        raise NotImplementedError(
            "decode is not yet implemented in %r" % type(self)
        )


def is_valid_mpm_plugin(mod: ModuleType) -> bool:
    """
    Return True if the module in *mod* is usable as MPM plugin
    """
    return all(
        [
            hasattr(mod, "create"),
            hasattr(mod, "IDENTIFIER"),
        ]
    )


def create(
    identifier: int,
    transport_handler: Callable[[bytes], Awaitable[bytes]],
    lcd: Dict[str, Any],
) -> MessageProcessingModel[Any, Any]:
    """
    Creates a new instance of a message-processing-model

    :param identifier: The IANA ID for the message-processing model
    :param transport_handler: A callable that is responsible to send data to
        the network. It will be called with bytes to be sent, and should
        return bytes. It should by itself already be aware of *where* to send
        them (the only input it gets is the payload). This is used in situations
        where the message-processing model needs to communicate with the device
        *outside* of normal data-exchange message, like the initial discovery
        message in the SNMPv3 protocol.
    :param lcd: A "local configuration directory" which is dynamically
        updated with "discovery" data if required.
    :returns: A new Message Processing Model instance


    >>> async def packet_handler(data: bytes) -> bytes:
    ...     loop = get_event_loop()
    ...     _, protocol = await loop.create_datagram_endpoint(
    ...         lambda: SNMPClientProtocol(packet, loop),
    ...         remote_addr=("192.0.2.1", 161),
    ...     )
    ...     response = await protocol.get_data(timeout)
    ...     return response
    >>> create(3, packet_handler, {})  # doctest: +ELLIPSIS
    <puresnmp_plugins.mpm.v3.V3MPM object ...>
    """
    # See https://tools.ietf.org/html/rfc3412#section-4.1.1
    namespace = "puresnmp_plugins.mpm"
    loader = Loader(namespace, is_valid_mpm_plugin)
    result = loader.create(identifier)
    if not result:
        raise UnknownMessageProcessingModel(
            namespace,
            identifier,
            sorted(loader.discovered_plugins.keys()),
        )
    return result.create(transport_handler, lcd)  # type: ignore



"""
This module contains variour type aliases for type checking
"""
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address
from typing import TypeVar, Union

#: This is a type variable helps to properly hint functions which take in a
#: "puresnmp" type and return a pythonic type
TWrappedPyType = TypeVar("TWrappedPyType")

TAnyIp = Union[IPv4Address, IPv6Address]


@dataclass(frozen=True)
class SocketInfo:
    """
    A simple tuple containing an IP address and port number
    """

    address: str
    port: int

    def __str__(self) -> str:
        return f"{self.address}:{self.port}"


@dataclass(frozen=True)
class SocketResponse:
    """
    Metadata for socket responses
    """

    data: bytes
    info: SocketInfo



"""
This module implements community based security model for SNMP
"""
from typing import Any, Dict

from x690.types import Integer, OctetString, Sequence

from puresnmp.credentials import V2C, Credentials
from puresnmp.exc import SnmpError
from puresnmp.pdu import PDU
from puresnmp.plugins.security import SecurityModel

IDENTIFIER = 2


class SNMPv2cSecurityModel(SecurityModel[PDU, Sequence]):
    """
    Implementation of the security model for community based SNMPv2 messages
    """

    def generate_request_message(
        self,
        message: PDU,
        security_engine_id: bytes,
        credentials: Credentials,
    ) -> Sequence:
        if not isinstance(credentials, V2C):
            raise SnmpError(
                "Credentials for the SNMPv2c security model must be "
                "V2C instances!"
            )
        packet = Sequence(
            [Integer(1), OctetString(credentials.community), message]
        )
        return packet

    def process_incoming_message(
        self,
        message: Sequence,
        credentials: Credentials,
    ) -> PDU:
        proto_version, community, pdu = message
        if not isinstance(credentials, V2C):
            raise SnmpError(
                "Credentials for the SNMPv2c security model must be "
                "V2C instances!"
            )
        if proto_version.pythonize() != 1:
            raise SnmpError(
                "Incoming SNMP message is not supported by the SNMPv2c "
                "security model. Most likely the device is not talking "
                "SNMPv2c but rather a different SNMP version."
            )
        if community.pythonize() != credentials.community.encode("ascii"):
            raise SnmpError("Mismatching community in response mesasge!")

        return pdu  # type: ignore


def create(local_config: Dict[bytes, Dict[str, Any]]) -> SNMPv2cSecurityModel:
    """
    Creates a new instance of the SNMPv2 community-based security model
    """
    return SNMPv2cSecurityModel(local_config)



"""
This module contains simple "abstract data types" which are used for both
type-hinting and to keep the code more expressive & readable.
"""
import sys
from dataclasses import dataclass
from textwrap import indent
from typing import Any, Type, TypeVar, Union, cast

from x690 import decode
from x690.types import Integer, OctetString, Sequence, X690Type

from puresnmp.pdu import PDU

if sys.stdout.isatty():
    # Add some colour for TTYs
    INDENT_STRING = " \u001b[38;5;22m│\u001b[0m "
else:
    INDENT_STRING = " │ "


TMessageType = TypeVar("TMessageType", bound="Message")


@dataclass(frozen=True)
class V3Flags:
    """
    This class represents the SNMP message flags.

    At the time of this writing, this only contains the flags auth, priv and
    reportable. The remaining bytes are (as of this state) reserved by the SNMP
    standard.
    """

    #: Whether the PDU exchange requires authentication
    auth: bool = False
    #: Whether the PDU exchange requires privacy (encryption)
    priv: bool = False
    #: Whether the exchanged PDU expects a response
    reportable: bool = False

    @staticmethod
    def decode(blob: OctetString) -> "V3Flags":
        """
        Converts an OctetString instance into a more pythonic instance

        >>> V3Flags.decode(OctetString(b"\\x00"))
        V3Flags(auth=False, priv=False, reportable=False)
        >>> V3Flags.decode(OctetString(b"\\x07"))
        V3Flags(auth=True, priv=True, reportable=True)
        """
        # TODO: Other "decode" methods use "bytes" as input value -> consistency
        flags = int.from_bytes(blob.pythonize(), "big")
        reportable = bool(flags & 0b100)
        priv = bool(flags & 0b010)
        auth = bool(flags & 0b001)
        return V3Flags(auth, priv, reportable)

    def __bytes__(self) -> bytes:
        value = 0
        value |= int(self.reportable) << 2
        value |= int(self.priv) << 1
        value |= int(self.auth)
        return bytes([value])


@dataclass(frozen=True)
class HeaderData:
    """
    Header information for an SNMPv3 message
    """

    #: ID to match responses with requests
    message_id: int
    #: Maximum number of bytes accepted by the SNMP engine
    message_max_size: int
    #: Additional boolean flags about the message
    flags: V3Flags
    #: The security model required to process this message (identified by a
    #: IANA registered ID). See :py:mod:`puresnmp.security`
    security_model: int

    # TODO: explode "flags" into its components

    def as_snmp_type(self) -> Sequence:
        """
        Convert this header-data object into an x.690 Sequence
        """
        return Sequence(
            [
                Integer(self.message_id),
                Integer(self.message_max_size),
                OctetString(bytes(self.flags)),
                Integer(self.security_model),
            ]
        )

    def __bytes__(self) -> bytes:
        return bytes(self.as_snmp_type())

    def pretty(self, depth: int = 0) -> str:
        """
        Return a prettyfied string of this object.

        This prettyfied string is useful for debugging and introspection.

        :param depth: How many levels of indentation to apply. This is used
            internally for nested structures.
        """
        lines = []
        lines.append("Global Data")
        lines.append(f"{INDENT_STRING}Message ID: {self.message_id}")
        lines.append(
            f"{INDENT_STRING}Message max size: {self.message_max_size}"
        )
        lines.append(f"{INDENT_STRING}Flags: {self.flags}")
        lines.append(f"{INDENT_STRING}Security Model: {self.security_model}")
        return indent("\n".join(lines), INDENT_STRING * depth)


@dataclass(frozen=True)
class ScopedPDU:
    """
    A data-structure wrapping the different values from a SNMPv3 "Scoped PDU"
    """

    context_engine_id: OctetString
    context_name: OctetString

    #: A PDU-structure as defined in SNMPv2
    data: PDU

    def __bytes__(self) -> bytes:
        return bytes(self.as_snmp_type())

    @staticmethod
    def decode(data: bytes, slc: slice = slice(None)) -> "ScopedPDU":
        """
        Create a ScopedPDU instance from an x.690 bytes object

        :param bytes: the bytes object which contains the PDU.
        :param slc: The slice at which the object is located. If left to the
            default, it will assume that the object is located in the beginning
            of the bytes object.

        >>> from puresnmp.pdu import GetRequest, PDUContent, VarBind
        >>> from x690.types import ObjectIdentifier as OID
        >>> pdu = ScopedPDU(
        ...     OctetString(b"engine"),
        ...     OctetString(b"context"),
        ...     GetRequest(PDUContent(42, [])),
        ... )
        >>> result = ScopedPDU.decode(bytes(pdu))
        >>> pdu == result
        True
        """
        sequence, _ = decode(
            data,
            start_index=slc.start or 0,
            enforce_type=Sequence,
            strict=False,
        )
        engine_id = cast(OctetString, sequence[0])
        cname = cast(OctetString, sequence[1])
        pdu = cast(PDU, sequence[2])
        output = ScopedPDU(
            context_engine_id=engine_id,
            context_name=cname,
            data=pdu,
        )
        return output

    def as_snmp_type(self) -> Sequence:
        """
        Convert this message into an x.690 Sequence
        """
        return Sequence(
            [
                self.context_engine_id,
                self.context_name,
                self.data,
            ]
        )

    def pretty(self, depth: int = 0) -> str:
        """
        Return a prettyfied string of this object.

        This prettyfied string is useful for debugging and introspection.

        :param depth: How many levels of indentation to apply. This is used
            internally for nested structures.
        """
        ids = INDENT_STRING
        lines = []
        lines.append("Scoped PDU")
        lines.append(
            f"{ids}Context Engine ID: {self.context_engine_id.value!r}"
        )
        lines.append(f"{ids}Context Name: {self.context_name.value!r}")
        lines.extend(self.data.pretty(1).splitlines())
        return indent("\n".join(lines), INDENT_STRING * depth)


@dataclass(frozen=True)
class Message:
    """
    A complete SNMP message including all information needed by :rfc:`3411`

    Such a message contains an "old-style" PDU (as was defined in SNMPv2)
    plus some meta-data. The meta-data consists of a version identifier to
    identify the proper "message-processing-model" using a IANA registered
    value.

    .. seealso::
        `RFC definition <https://tools.ietf.org/html/rfc3412#section-6>`_
            The MIB definition of an SNMPv3 message
        `Example Message Format <http://www.tcpipguide.com/free/t_SNMPVersion3SNMPv3MessageFormat.htm>`_
            A simple visual representation of the byte-structure of an SNMPv3 message
        `IANA Message Processing Models <https://www.iana.org/assignments/snmp-number-spaces/snmp-number-spaces.xml#snmp-number-spaces-2>`_
            Oficially registered message "version" identifiers
    """

    #: The IANA version identifier
    version: Integer
    #: Additional information wrapping the old-style PDU
    header: HeaderData
    #: Additional data needed to authenticate & en/decrypt the message
    security_parameters: bytes
    #: The "old-style" PDU (either plain or encrypted)
    scoped_pdu: Union[OctetString, ScopedPDU]

    def __bytes__(self) -> bytes:
        spdu: X690Type[Any]
        if isinstance(self.scoped_pdu, ScopedPDU):
            spdu = self.scoped_pdu.as_snmp_type()
        else:
            spdu = self.scoped_pdu

        output = bytes(
            Sequence(
                [
                    self.version,
                    self.header.as_snmp_type(),
                    OctetString(self.security_parameters),
                    spdu,
                ]
            )
        )
        return output

    @classmethod
    def from_sequence(cls: Type[TMessageType], seq: Sequence) -> TMessageType:
        """
        Construct a Message instance from an X.690 Sequence
        """
        version = cast(Integer, seq[0])
        header = cast(Sequence, seq[1])
        security_parameters = cast(OctetString, seq[2]).value

        msg_id = cast(Integer, header[0])
        msg_max_size = cast(Integer, header[1])
        security_level = V3Flags.decode(cast(OctetString, header[2]))
        security_model_id = cast(Integer, header[3]).pythonize()

        if security_level.priv:
            payload: Union[OctetString, ScopedPDU] = cast(OctetString, seq[3])
        else:
            scoped_pdu = cast(Sequence, seq[3])
            engine_id = cast(OctetString, scoped_pdu[0])
            context_name = cast(OctetString, scoped_pdu[1])
            pdu = cast(PDU, scoped_pdu[2])
            payload = ScopedPDU(engine_id, context_name, pdu)

        output = cls(
            version,
            HeaderData(
                msg_id.pythonize(),
                msg_max_size.pythonize(),
                security_level,
                security_model_id,
            ),
            security_parameters,
            payload,
        )

        return output

    @staticmethod
    def decode(data: bytes) -> Union["PlainMessage", "EncryptedMessage"]:
        """
        Construct a new SNMPv3 message from a bytes object

        >>> from puresnmp.pdu import GetRequest, PDUContent
        >>> from x690.types import ObjectIdentifier as OID
        >>> pdu = ScopedPDU(
        ...     OctetString(b"engine"),
        ...     OctetString(b"context"),
        ...     GetRequest(PDUContent(42, [])),
        ... )
        >>> msg = PlainMessage(
        ...     version=Integer(3),
        ...     header=HeaderData(42, 65000, V3Flags(False, False, False), 3),
        ...     security_parameters=b"",
        ...     scoped_pdu=pdu
        ... )
        >>> result = Message.decode(bytes(msg))
        >>> result == msg
        True
        """

        message, _ = decode(data, enforce_type=Sequence)
        cls = (
            EncryptedMessage
            if isinstance(message[3], OctetString)
            else PlainMessage
        )
        return cls.from_sequence(message)  # type: ignore

    def pretty(self, depth: int = 0) -> str:
        """
        Return a prettyfied string of this object.

        This prettyfied string is useful for debugging and introspection.

        :param depth: How many levels of indentation to apply. This is used
            internally for nested structures.
        """
        lines = []

        lines.append(f"SNMP Message (version-identifier={self.version})")
        lines.extend(self.header.pretty(depth + 1).splitlines())
        lines.append(indent("Security Parameters", INDENT_STRING * (depth + 1)))
        lines.extend(
            OctetString(self.security_parameters).pretty(depth + 2).splitlines()
        )
        if isinstance(self.scoped_pdu, bytes):
            lines.append(
                indent("Scoped PDU (encrypted)", INDENT_STRING * (depth + 1))
            )
            lines.extend(
                OctetString(self.scoped_pdu).pretty(depth + 2).splitlines()
            )
        else:
            lines.extend(self.scoped_pdu.pretty(depth + 1).splitlines())
        return indent("\n".join(lines), INDENT_STRING * depth)


class PlainMessage(Message):
    """
    A message whose PDU is not encrypted
    """

    version: Integer
    header: HeaderData
    security_parameters: bytes
    scoped_pdu: ScopedPDU


class EncryptedMessage(Message):
    """
    A message whose PDU is encrypted
    """

    version: Integer
    header: HeaderData
    security_parameters: bytes
    scoped_pdu: OctetString



"""
This module provides a plugin architecture for authentication methods.

Each authentication plugin can be distributed as separate package by providing
modules inside the namespace-package ``puresnmp_plugins.auth``. Note that in
order to be a valid namespace-package, such a package *must not* have a
``__init__.py`` file!

Example folder-structure for a privacy plugin:

.. code-block:: text

    my-auth-plugin/
     +- setup.py (or pyproject.toml)
     +- puresnmp/
         +- auth/
             +- mymodule.py
             +- myothermodule.py

Note that there is no ``__init__.py`` file!

In order for modules to be detected as plugin, they must follow the following
rules:

* Have a function ``authenticate_outgoing_message`` implementing the
  :py:meth:`puresnmp_plugins.auth.TAuth.authenticate_outgoing_message`
  protocol.
* Have a function ``authenticate_incoming_message`` implementing the
  :py:meth:`puresnmp_plugins.auth.TAuth.authenticate_incoming_message`
  protocol.
* Contain a string-variable ``IDENTIFIER``. This variable should be
  user-friendly and is used to uniquely identify this authentication module.
* Contain a int-variable ``IANA_ID``. This variable should have a value from
  IANA registered authentication protocols and is used to avoid duplicate plugin
  registrations. See
  https://www.iana.org/assignments/snmp-number-spaces/snmp-number-spaces.xhtml
  and :rfc:`3411`
"""
import importlib
from threading import Lock
from types import ModuleType
from typing import Dict

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


from puresnmp.exc import MissingPlugin, UnknownAuthModel, UnknownPrivacyModel
from puresnmp.plugins.pluginbase import Loader
from puresnmp.util import iter_namespace


class TAuth(Protocol):
    """
    Protocol describing the authentication API
    """

    def authenticate_outgoing_message(
        self, auth_key: bytes, data: bytes, engine_id: bytes
    ) -> bytes:  # pragma: no cover
        """
        Calculate a digest for an outgoing message

        This will take the raw bytes of an outgoing message and derive a
        digest based on the authentication key and the receipient engine ID
        """
        ...

    def authenticate_incoming_message(
        self,
        auth_key: bytes,
        data: bytes,
        received_digest: bytes,
        engine_id: bytes,
    ) -> bool:  # pragma: no cover
        """
        Determine whether a message is authentic.

        This will take the raw bytes of an incoming message and derive a
        digest based on the authentication key and the receipient engine ID.
        The digest is compared to an expected digest. If they differ, this
        function will return False.
        """
        ...


def is_valid_auth_mod(mod: ModuleType) -> bool:
    return all(
        [
            hasattr(mod, "authenticate_incoming_message"),
            hasattr(mod, "authenticate_outgoing_message"),
            hasattr(mod, "IDENTIFIER"),
            hasattr(mod, "IANA_ID"),
        ]
    )


def create(identifier: str) -> TAuth:
    """
    Return an instance of the given authentication module by identifier.

    This looks up the module by "IDENTIFIER" as specified in the given plugin.

    :param name: The identifier of the module
    :raises puresnmp.exc.MissingPlugin: If no module with the given name is
        found
    """
    namespace = "puresnmp_plugins.auth"
    loader = Loader(namespace, is_valid_auth_mod)
    result = loader.create(identifier)
    if not result:
        raise UnknownAuthModel(
            namespace,
            identifier,
            sorted(loader.discovered_plugins.keys()),
        )
    return result  # type: ignore



from typing import Any, Awaitable, Callable, Dict

from puresnmp.plugins.mpm import AbstractEncodingResult, MessageProcessingModel

IDENTIFIER = 2


class V2XMPM(MessageProcessingModel[AbstractEncodingResult, Any]):
    """
    Message Processing Model for SNMP v2*
    """


def create(
    transport_handler: Callable[[bytes], Awaitable[bytes]],
    lcd: Dict[str, Any],
) -> "MessageProcessingModel[AbstractEncodingResult, Any]":
    """
    Creates a new instance for V1 message processing
    """
    return V2XMPM(transport_handler, lcd)
